from __future__ import annotations

import json
import sqlite3
import time
import uuid
from typing import Any

from x8._common.sqlite_provider import SQLiteProvider
from x8.core import Context, DataModel, NCall, Operation, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError
from x8.core.time import Time

from ._models import (
    DEFAULT_SUBSCRIPTION_NAME,
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingMode,
    QueueConfig,
    QueueInfo,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)
from ._operation import MessagingOperation
from ._operation_parser import MessagingOperationParser

DEFAULT_TTL = 7 * 24 * 60 * 60
DEFAULT_VISIBILITY_TIMEOUT = 30


class SQLiteBase(SQLiteProvider):
    mode: MessagingMode

    queue: str | None
    topic: str | None
    subscription: str | None

    database: str
    message_table: str
    metadata_table: str
    poll_interval: float
    nparams: dict[str, Any]

    _client: Any
    _client_helper: Any
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _topic_config_cache: dict[str, TopicConfig] = {}
    _subscription_config_cache: dict[
        str, dict[str, SubscriptionConfig | None]
    ] = {}

    def __init__(
        self,
        mode: MessagingMode = MessagingMode.QUEUE,
        database: str = ":memory:",
        queue: str | None = None,
        topic: str | None = None,
        subscription: str | None = None,
        message_table: str = "message",
        metadata_table: str = "metadata",
        poll_interval: float = 0.5,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            mode:
                Messaging mode.
            database:
                SQLite database. Defaults to ":memory:".
            queue:
                Queue name.
            topic:
                Topic name.
            subscription:
                Subscription name.
            message_table:
                SQLite table name for messages.
            metadata_table:
                SQLite table name for metadata.
            poll_interval:
                Poll interval in seconds. Defaults to 0.5.
            nparams:
                Native parameters to SQLite client.
        """
        self.mode = mode
        self.database = database
        self.queue = queue
        self.topic = topic
        self.subscription = subscription
        self.message_table = message_table
        self.metadata_table = metadata_table
        self.poll_interval = poll_interval
        self.nparams = nparams

        self._client = None
        self._client_helper = None
        self._op_converter = OperationConverter(
            self.message_table,
            self.metadata_table,
        )
        self._result_converter = ResultConverter()
        self._topic_config_cache = dict()
        self._subscription_config_cache = dict()
        if self.mode == MessagingMode.QUEUE:
            self.topic = self.queue
            self.subscription = DEFAULT_SUBSCRIPTION_NAME
        super().__init__(**kwargs)

    def __setup__(
        self,
        context: Context | None = None,
    ) -> None:
        if self._client is not None:
            return

        self._init_folder(self.database)
        self._client = sqlite3.connect(
            database=self.database,
            check_same_thread=False,
            **self.nparams,
        )
        self._client_helper = ClientHelper(
            self._client,
            self._op_converter,
            self._result_converter,
            self.mode,
            self.poll_interval,
            self._subscription_config_cache,
        )
        self._create_tables_if_needed()

    def _create_tables_if_needed(self, **kwargs: Any) -> None:
        cursor = self._client.cursor()
        rows = cursor.execute(
            """SELECT name FROM sqlite_master WHERE type='table'
                AND name=?""",
            (self.message_table,),
        ).fetchall()
        if len(rows) == 0:
            cursor.execute(
                f"""
                CREATE TABLE {self.message_table} (
                    id TEXT,
                    topic TEXT,
                    subscription TEXT,
                    value BLOB,
                    metadata TEXT,
                    message_id TEXT,
                    group_id TEXT,
                    content_type TEXT,
                    enqueued_time REAL,
                    delivery_count INTEGER,
                    lock_until_time REAL,
                    lock_token TEXT,
                    PRIMARY KEY (id, topic, subscription)
                )
                """
            )
        rows = cursor.execute(
            """SELECT name FROM sqlite_master WHERE type='table'
                AND name=?""",
            (self.metadata_table,),
        ).fetchall()
        if len(rows) == 0:
            cursor.execute(
                f"""
                CREATE TABLE {self.metadata_table} (
                    topic TEXT,
                    subscription TEXT,
                    config TEXT,
                    PRIMARY KEY (topic, subscription)
                )
                """
            )
        self._client.commit()
        cursor.close()

    def _get_topic_name(self, op_parser: MessagingOperationParser) -> str:
        if self.mode == MessagingMode.PUBSUB:
            topic_name = op_parser.get_topic()
            topic_name = topic_name or self.topic or self.__component__.topic
            if topic_name is None:
                raise BadRequestError(
                    "Topic name is required for this operation.",
                )
        else:
            topic_name = op_parser.get_queue()
            topic_name = topic_name or self.queue or self.__component__.queue
            if topic_name is None:
                raise BadRequestError(
                    "Queue name is required for this operation.",
                )
        return topic_name

    def _get_subscription_name(
        self, op_parser: MessagingOperationParser
    ) -> str:
        subscription_name = op_parser.get_subscription()
        subscription_name = (
            subscription_name
            or self.subscription
            or self.__component__.subscription
        )
        if subscription_name is None:
            raise BadRequestError(
                "Subscription name is required for this operation.",
            )
        return subscription_name

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._client_helper,
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        helper: ClientHelper,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        op_converter = self._op_converter
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            args = op_converter.convert_create_queue(
                self._get_topic_name(op_parser),
                op_parser.get_queue_config(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.transact,
                args,
                None,
                {sqlite3.IntegrityError: ConflictError},
            )
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            args = op_converter.convert_update_queue(
                self._get_topic_name(op_parser),
                op_parser.get_queue_config(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            args = op_converter.convert_drop_topic(
                self._get_topic_name(op_parser),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            args = op_converter.convert_has_topic(
                self._get_topic_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            args = {
                "queue": self._get_topic_name(op_parser),
            }
            call = NCall(
                helper.get_queue,
                args,
            )
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            args = op_converter.convert_list_topics()
            call = NCall(
                helper.execute,
                args,
            )
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            args = op_converter.convert_create_topic(
                self._get_topic_name(op_parser),
                op_parser.get_topic_config(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: ConflictError},
            )
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            args = op_converter.convert_update_topic(
                self._get_topic_name(op_parser),
                op_parser.get_topic_config(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            args = op_converter.convert_drop_topic(
                self._get_topic_name(op_parser),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            args = op_converter.convert_has_topic(
                self._get_topic_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
            }
            call = NCall(
                helper.get_topic,
                args,
            )
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            args = op_converter.convert_list_topics()
            call = NCall(
                helper.execute,
                args,
            )
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            args = op_converter.convert_create_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_subscription_config(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: ConflictError},
            )
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            args = op_converter.convert_update_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_subscription_config(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            args = op_converter.convert_drop_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            args = op_converter.convert_has_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
            }
            call = NCall(
                helper.get_subscription,
                args,
            )
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            args = op_converter.convert_list_subscriptions(
                self._get_topic_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = {
                "topic": self._get_topic_name(op_parser),
                "value": op_parser.get_value(),
                "metadata": op_parser.get_metadata(),
                "properties": op_parser.get_properties(),
                "config": op_parser.get_put_config(),
            }
            call = NCall(
                helper.put,
                args,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = {
                "topic": self._get_topic_name(op_parser),
                "batch": op_parser.get_batch(),
            }
            call = NCall(
                helper.batch,
                args,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "config": op_parser.get_pull_config(),
            }
            call = NCall(
                helper.pull,
                args,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = op_converter.convert_ack(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_key(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = op_converter.convert_nack(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_key(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "key": op_parser.get_key(),
                "timeout": op_parser.get_timeout(),
            }
            call = NCall(
                helper.extend,
                args,
            )
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            args = op_converter.convert_purge(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {
                "topic": op_parser.get_topic() or op_parser.get_queue(),
            }
            call = NCall(helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = self._result_converter
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            result = None
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            if nresult == 0:
                raise NotFoundError(
                    "Queue not found.",
                )
            result = None
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            where_exists = op_parser.get_where_exists()
            if nresult == 0 and where_exists is True:
                raise NotFoundError
            result = None
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            return len(nresult) > 0
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            result = nresult
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            result = result_converter.convert_list_topics(nresult)
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            result = None
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            if nresult == 0:
                raise NotFoundError(
                    "Topic not found.",
                )
            result = None
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            where_exists = op_parser.get_where_exists()
            if nresult == 0 and where_exists is True:
                raise NotFoundError
            result = None
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            return len(nresult) > 0
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            result = nresult
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            result = result_converter.convert_list_topics(nresult)
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            result = None
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            if nresult == 0:
                raise NotFoundError(
                    "Subscription not found.",
                )
            result = None
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            where_exists = op_parser.get_where_exists()
            if nresult == 0 and where_exists is True:
                raise NotFoundError
            result = None
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            return len(nresult) > 0
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            result = nresult
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            result = result_converter.convert_list_subscriptions(nresult)
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = nresult
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            if nresult == 0:
                raise BadRequestError(
                    "Message not found or already acknowledged.",
                )
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            if nresult == 0:
                raise BadRequestError(
                    "Message not found or already acknowledged.",
                )
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            if nresult == 0:
                raise BadRequestError(
                    "Message not found or already acknowledged.",
                )
            result = MessageItem(key=op_parser.get_key())
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            pass
        return result


class ClientHelper:
    client: Any
    op_converter: OperationConverter
    result_converter: ResultConverter
    mode: MessagingMode
    poll_interval: float
    subscription_config_cache: dict[str, dict[str, SubscriptionConfig | None]]

    def __init__(
        self,
        client: Any,
        op_converter: OperationConverter,
        result_converter: ResultConverter,
        mode: MessagingMode,
        poll_interval: float,
        subscription_config_cache: dict[
            str, dict[str, SubscriptionConfig | None]
        ],
    ):
        self.client = client
        self.op_converter = op_converter
        self.result_converter = result_converter
        self.mode = mode
        self.poll_interval = poll_interval
        self.subscription_config_cache = subscription_config_cache

    def _get_subscription_config(
        self,
        topic_name: str,
        subscription_name: str,
    ) -> SubscriptionConfig | None:
        if (
            topic_name in self.subscription_config_cache
            and subscription_name in self.subscription_config_cache[topic_name]
        ):
            return self.subscription_config_cache[topic_name][
                subscription_name
            ]
        args = self.op_converter.convert_get_subscription_config(
            topic_name,
            subscription_name,
        )
        nresult = self.execute(
            args["query"],
            args["params"],
            fetchall=True,
        )
        if nresult is None or len(nresult) == 0:
            raise BadRequestError(
                f"Subscription {subscription_name} not found.",
            )
        if nresult[0][0] is None:
            config = None
        else:
            config = SubscriptionConfig.from_dict(json.loads(nresult[0][0]))
        if topic_name not in self.subscription_config_cache:
            self.subscription_config_cache[topic_name] = {}
        self.subscription_config_cache[topic_name][subscription_name] = config
        return config

    def get_queue(self, queue: str) -> QueueInfo:
        sinfo = self.get_subscription(queue, DEFAULT_SUBSCRIPTION_NAME)
        return QueueInfo(
            name=queue,
            active_message_count=sinfo.active_message_count,
            inflight_message_count=sinfo.inflight_message_count,
            scheduled_message_count=sinfo.scheduled_message_count,
            config=(
                QueueConfig.from_dict(sinfo.config.to_dict())
                if sinfo.config
                else None
            ),
            nref=queue,
        )

    def get_topic(self, topic: str) -> TopicInfo:
        args = self.op_converter.convert_get_topic_config(topic)
        nresult = self.execute(args["query"], args["params"])
        if nresult is None:
            raise NotFoundError
        config_json = nresult[0]
        args = self.op_converter.convert_get_subscription_count(topic)
        subscription_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        return TopicInfo(
            name=topic,
            subscription_count=subscription_count - 1,
            config=(
                TopicConfig.from_dict(json.loads(config_json))
                if config_json
                else None
            ),
            nref=topic,
        )

    def get_subscription(
        self, topic: str, subscription: str
    ) -> SubscriptionInfo:
        args = self.op_converter.convert_get_subscription_config(
            topic, subscription
        )
        nresult = self.execute(args["query"], args["params"])
        if nresult is None:
            raise NotFoundError
        config_json = nresult[0]
        args = self.op_converter.convert_get_active_message_count(
            topic, subscription
        )
        active_message_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        args = self.op_converter.convert_get_inflight_message_count(
            topic, subscription
        )
        inflight_message_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        args = self.op_converter.convert_get_scheduled_message_count(
            topic, subscription
        )
        scheduled_message_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        return SubscriptionInfo(
            name=subscription,
            topic=topic,
            active_message_count=active_message_count,
            inflight_message_count=inflight_message_count,
            scheduled_message_count=scheduled_message_count,
            config=(
                SubscriptionConfig.from_dict(json.loads(config_json))
                if config_json
                else None
            ),
            nref=subscription,
        )

    def put(
        self,
        topic: str,
        value: MessageValueType,
        metadata: dict | None = None,
        properties: MessageProperties | None = None,
        config: MessagePutConfig | None = None,
    ) -> None:
        args = self.op_converter.convert_list_subscriptions(topic)
        nresult = self.client.execute(args["query"], args["params"]).fetchall()
        subscriptions = self.result_converter.convert_list_subscriptions(
            nresult
        )
        if len(subscriptions) == 0:
            if self.mode == MessagingMode.QUEUE:
                raise BadRequestError(
                    f"Queue {topic} not found.",
                )
        if properties and properties.message_id:
            id = properties.message_id
        else:
            id = str(uuid.uuid4())
        enqueue_time = Time.now()
        if config and config.delay:
            enqueue_time += config.delay
        ops: list = []
        for subscription in subscriptions:
            args = self.op_converter.convert_put(
                id,
                topic,
                subscription,
                enqueue_time,
                value,
                metadata,
                properties,
            )
            ops.append(args)
        self.transact(ops=ops)

    def batch(
        self,
        topic: str,
        batch: MessageBatch,
    ) -> None:
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            self.put(
                topic=topic,
                value=op_parser.get_value(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                config=op_parser.get_put_config(),
            )

    def pull(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None = None,
    ) -> list[MessageItem]:
        subscription_config = self._get_subscription_config(
            topic,
            subscription,
        )
        result: list[MessageItem] = []
        start_time = Time.now()
        while True:
            now = Time.now()
            found = False
            with self.client:
                args = self.op_converter.convert_pull(
                    topic,
                    subscription,
                    subscription_config=subscription_config,
                )
                nresult = self.client.execute(
                    args["query"], args["params"]
                ).fetchone()
                if nresult is not None:
                    lock_token = str(uuid.uuid4())
                    message = self.result_converter.convert_pull(nresult)
                    if message.key is None or message.key.id is None:
                        raise BadRequestError(
                            "Message key is missing.",
                        )
                    lock_duration = (
                        config.visibility_timeout if config else None
                    )
                    lock_duration = lock_duration or (
                        subscription_config.visibility_timeout
                        if subscription_config
                        and subscription_config.visibility_timeout
                        else DEFAULT_VISIBILITY_TIMEOUT
                    )
                    lock_until_time = now + lock_duration
                    args = self.op_converter.convert_lock(
                        message.key.id,
                        topic,
                        subscription,
                        lock_until_time,
                        lock_token,
                        message.key.nref,
                    )
                    updated = self.client.execute(
                        args["query"], args["params"]
                    ).rowcount
                    if updated == 1:
                        message.key.nref = lock_token
                        result.append(message)
                        found = True
            if config and config.max_wait_time:
                if now - start_time > config.max_wait_time:
                    break
            if config and config.max_count:
                if len(result) >= config.max_count:
                    break
            if not config or not config.max_count:
                if len(result) > 0:
                    break
            if not found:
                time.sleep(self.poll_interval)
        return result

    def extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None = None,
    ) -> None:
        subscription_config = self._get_subscription_config(
            topic, subscription
        )
        args = self.op_converter.convert_extend(
            topic,
            subscription,
            key,
            timeout,
            subscription_config,
        )
        self.client.execute(args["query"], args["params"])

    def execute(
        self,
        query: str,
        params: tuple | None = None,
        rowcount: bool = False,
        fetchall: bool = False,
    ) -> Any:
        cursor = self.client.cursor()
        try:
            if params:
                cursor.execute(query, params)
            else:
                cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return cursor.fetchall()
            return cursor.fetchone()
        finally:
            self.client.commit()
            cursor.close()

    def transact(self, ops: list) -> Any:
        result = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                if "params" in op:
                    cursor.execute(op["query"], op["params"])
                else:
                    cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    if cursor.rowcount == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(cursor.rowcount)
                else:
                    res = cursor.fetchone()
                    if res is None or len(res) == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(res)
            self.client.commit()
        finally:
            cursor.close()
        return result

    def close(self, topic: str | None) -> Any:
        if topic is not None:
            return
        self.client.close()


class OperationConverter:
    message_table: str
    metadata_table: str

    META_SUBSCRIPTION_NAME = "#"

    def __init__(
        self,
        message_table: str,
        metadata_table: str,
    ):
        self.message_table = message_table
        self.metadata_table = metadata_table

    def convert_get_active_message_count(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        current_time = Time.now()
        query = f"""
            SELECT COUNT(*)
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_until_time < ?
            AND enqueued_time <= ?
            """
        params = (
            topic,
            subscription,
            current_time,
            current_time,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_inflight_message_count(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        current_time = Time.now()
        query = f"""
            SELECT COUNT(*)
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_until_time >= ?
            """
        params = (
            topic,
            subscription,
            current_time,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_scheduled_message_count(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        current_time = Time.now()
        query = f"""
            SELECT COUNT(*)
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND enqueued_time > ?
            """
        params = (
            topic,
            subscription,
            current_time,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_subscription_count(
        self,
        topic: str,
    ) -> dict:
        query = f"""
            SELECT COUNT(*)
            FROM {self.metadata_table}
            WHERE topic = ?
            """
        params = (topic,)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_topic_config(
        self,
        topic: str,
    ) -> dict:
        query = f"""
            SELECT config
            FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, OperationConverter.META_SUBSCRIPTION_NAME)
        return {
            "query": query,
            "params": params,
        }

    def convert_get_subscription_config(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        query = f"""
            SELECT config
            FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
        }

    def convert_create_queue(
        self,
        queue: str,
        config: QueueConfig | None,
        where_exists: bool | None,
    ) -> dict:
        ops: list = []
        args = self.convert_create_topic(
            queue,
            None,
            where_exists,
        )
        ops.append(args)
        args = self.convert_create_subscription(
            queue,
            DEFAULT_SUBSCRIPTION_NAME,
            config,
            where_exists,
        )
        ops.append(args)
        return {"ops": ops}

    def convert_update_queue(
        self,
        queue: str,
        config: QueueConfig | None,
    ) -> dict:
        return self.convert_update_subscription(
            queue,
            DEFAULT_SUBSCRIPTION_NAME,
            config,
        )

    def convert_create_topic(
        self,
        topic: str,
        config: QueueConfig | TopicConfig | None,
        where_exists: bool | None,
    ) -> dict:
        if where_exists is False:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                """
        else:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                ON CONFLICT (topic, subscription)
                DO UPDATE SET
                config = EXCLUDED.config
                """
        params = (
            topic,
            OperationConverter.META_SUBSCRIPTION_NAME,
            config.to_json() if config else None,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_update_topic(
        self,
        topic: str,
        config: TopicConfig | None,
    ) -> dict:
        query = f"""
            UPDATE {self.metadata_table}
            SET config = ?
            WHERE topic = ? AND subscription = ?
            """
        params = (
            config.to_json() if config else None,
            topic,
            OperationConverter.META_SUBSCRIPTION_NAME,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_drop_topic(
        self,
        topic: str,
        where_exists: bool | None,
    ) -> dict:
        query = f"""
            DELETE FROM {self.metadata_table}
            WHERE topic = ?
            """
        params = (topic,)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_has_topic(
        self,
        topic: str,
    ) -> dict:
        query = f"""
            SELECT topic
            FROM {self.metadata_table}
            WHERE topic = ?
            """
        params = (topic,)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_list_topics(self) -> dict:
        query = f"""
            SELECT topic
            FROM {self.metadata_table}
            WHERE subscription = ?
            """
        params = (OperationConverter.META_SUBSCRIPTION_NAME,)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_create_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | QueueConfig | None,
        where_exists: bool | None,
    ) -> dict:
        if where_exists is False:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                """
        else:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                ON CONFLICT (topic, subscription)
                DO UPDATE SET
                config = EXCLUDED.config
                """
        params = (
            topic,
            subscription,
            config.to_json() if config else None,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_update_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | QueueConfig | None,
    ) -> dict:
        query = f"""
            UPDATE {self.metadata_table}
            SET config = ?
            WHERE topic = ? AND subscription = ?
            """
        params = (
            config.to_json() if config else None,
            topic,
            subscription,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_drop_subscription(
        self,
        topic: str,
        subscription: str,
        where_exists: bool | None,
    ) -> dict:
        query = f"""
            DELETE FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_has_subscription(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        query = f"""
            SELECT topic, subscription
            FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_list_subscriptions(self, topic: str) -> dict:
        query = f"""
            SELECT subscription
            FROM {self.metadata_table}
            WHERE topic = ?
            AND subscription != ?
            """
        params = (
            topic,
            OperationConverter.META_SUBSCRIPTION_NAME,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_put(
        self,
        id: str,
        topic: str,
        subscription: str,
        enqueue_time: float,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
    ) -> dict:
        message_id = None
        content_type = None
        group_id = None
        body: bytes | None = None
        if properties is not None:
            message_id = properties.message_id
            content_type = properties.content_type
            group_id = properties.group_id
        if isinstance(value, str):
            body = value.encode("utf-8")
            content_type = "text/plain"
        elif isinstance(value, bytes):
            body = value
            content_type = "application/octet-stream"
        elif isinstance(value, dict):
            body = json.dumps(value).encode("utf-8")
            content_type = "application/json"
        elif isinstance(value, DataModel):
            body = value.to_json().encode("utf-8")
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")
        query = f"""
            INSERT INTO {self.message_table}
            (id, topic, subscription, value, metadata,
            message_id, group_id, content_type,
            enqueued_time, delivery_count,
            lock_until_time, lock_token)
            VALUES (?, ?, ?, ?, ?,
            ?, ?, ?, ?, ?, ?, ?)
            """
        params = (
            id,
            topic,
            subscription,
            body,
            json.dumps(metadata) if metadata else None,
            message_id,
            group_id,
            content_type,
            enqueue_time,
            0,
            0,
            "",
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_pull(
        self,
        topic: str,
        subscription: str,
        subscription_config: SubscriptionConfig | None = None,
    ) -> dict:
        ttl = (
            subscription_config.ttl
            if subscription_config and subscription_config.ttl
            else DEFAULT_TTL
        )
        current_time = Time.now()
        query = f"""
            SELECT id, topic, subscription, value, metadata,
            message_id, group_id, content_type,
            enqueued_time, delivery_count,
            lock_until_time, lock_token
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_until_time <= ?
            AND enqueued_time <= ?
            AND enqueued_time >= ?
            ORDER BY enqueued_time
            LIMIT 1
            """
        params = (
            topic,
            subscription,
            current_time,
            current_time,
            current_time - ttl,
        )
        return {
            "query": query,
            "params": params,
        }

    def convert_lock(
        self,
        id: str,
        topic: str,
        subscription: str,
        lock_until_time: float,
        lock_token: str,
        previous_lock_token: str | None,
    ) -> dict:
        query = f"""
            UPDATE {self.message_table}
            SET lock_until_time = ?, lock_token = ?,
            delivery_count = delivery_count + 1
            WHERE topic = ? AND subscription = ?
            AND id = ? AND lock_token = ?
            """
        params = (
            lock_until_time,
            lock_token,
            topic,
            subscription,
            id,
            previous_lock_token,
        )
        return {
            "query": query,
            "params": params,
        }

    def convert_ack(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
    ) -> dict:
        query = f"""
            DELETE FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_token = ?
            """
        params = (topic, subscription, key.nref)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_nack(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
    ) -> dict:
        query = f"""
            UPDATE {self.message_table}
            SET lock_until_time = ?, lock_token = ?
            WHERE topic = ? AND subscription = ?
            AND lock_token = ?
            """
        params = (0, "", topic, subscription, key.nref)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None,
        subscription_config: SubscriptionConfig | None = None,
    ) -> dict:
        lock_duration = (
            subscription_config.visibility_timeout
            if subscription_config and subscription_config.visibility_timeout
            else DEFAULT_VISIBILITY_TIMEOUT
        )
        t = timeout or lock_duration
        t = Time.now() + t
        query = f"""
            UPDATE {self.message_table}
            SET lock_until_time = ?
            WHERE topic = ? AND subscription = ?
            AND lock_token = ?
            """
        params = (t, topic, subscription, key.nref)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_purge(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        query = f"""
            DELETE FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }


class ResultConverter:
    def __init__(self):
        pass

    def _convert_value(
        self,
        value: bytes,
        content_type: str | None,
    ) -> MessageValueType:
        if content_type == "text/plain":
            return value.decode("utf-8")
        elif content_type == "application/octet-stream":
            return value
        elif content_type == "application/json":
            return json.loads(value.decode("utf-8"))
        return value

    def convert_list_topics(self, nresult: Any) -> list[str]:
        topics: list[str] = []
        for row in nresult:
            topics.append(row[0])
        return topics

    def convert_list_subscriptions(self, nresult: Any) -> list[str]:
        subscriptions: list[str] = []
        for row in nresult:
            subscriptions.append(row[0])
        return subscriptions

    def convert_pull(
        self,
        nresult: Any,
    ) -> MessageItem:
        id = nresult[0]
        value = nresult[3]
        metadata = nresult[4]
        message_id = nresult[5]
        group_id = nresult[6]
        content_type = nresult[7]
        enqueued_time = nresult[8]
        delivery_count = nresult[9]
        lock_token = nresult[11]

        message = MessageItem(
            key=MessageKey(id=id, nref=lock_token),
            value=self._convert_value(
                value,
                content_type,
            ),
            metadata=json.loads(metadata) if metadata else None,
            properties=MessageProperties(
                message_id=message_id,
                group_id=group_id,
                content_type=content_type,
                enqueued_time=enqueued_time,
                delivery_count=delivery_count + 1,
            ),
        )
        return message



from typing import Any, Mapping

from openai import AsyncAzureOpenAI
from openai import AzureOpenAI as AzureOpenAIClient

from .openai_legacy import OpenAILegacy


class AzureOpenAI(OpenAILegacy):
    """Azure OpenAI provider using the Chat Completions API.

    This provider uses Azure's OpenAI Service with the Chat Completions API.
    It inherits from OpenAILegacy and configures Azure-specific client
    settings.
    """

    api_version: str | None
    azure_endpoint: str | None
    azure_deployment: str | None
    azure_ad_token: str | None
    organization: str | None

    def __init__(
        self,
        api_key: str | None = None,
        api_version: str | None = None,
        azure_endpoint: str | None = None,
        azure_deployment: str | None = None,
        azure_ad_token: str | None = None,
        organization: str | None = None,
        timeout: float | None = None,
        max_retries: int | None = None,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        model: str = "gpt-4o",
        max_tokens: int = 8192,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        """Initialize Azure OpenAI provider.

        Args:
            api_key:
                Azure OpenAI API key.
            api_version:
                Azure API version (e.g., "2024-12-01-preview").
            azure_endpoint:
                Azure endpoint URL
                (e.g., "https://your-resource.openai.azure.com/").
            azure_deployment:
                Azure deployment name. If not provided, uses model name.
            azure_ad_token:
                Azure Active Directory token for authentication.
            organization:
                OpenAI organization ID (optional for Azure).
            timeout:
                Timeout for client requests.
            max_retries:
                Maximum number of retries for failed requests.
            default_headers:
                Default headers to include in every request.
            default_query:
                Default query parameters to include in every request.
            model:
                Model deployment name to use for text generation.
            max_tokens:
                Default maximum tokens for responses.
            nparams:
                Native params for Azure OpenAI client.
        """
        self.api_version = api_version
        self.azure_endpoint = azure_endpoint
        self.azure_deployment = azure_deployment
        self.azure_ad_token = azure_ad_token

        # Call parent init
        # Note: we don't pass base_url since Azure uses azure_endpoint
        super().__init__(
            api_key=api_key,
            organization=organization,
            project=None,  # Azure doesn't use project
            base_url=None,  # Azure uses azure_endpoint instead
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            model=model,
            max_tokens=max_tokens,
            nparams=nparams,
            **kwargs,
        )

    def __setup__(self, context=None):
        if self._init:
            return

        client_kwargs = self._get_azure_client_params()
        self._client = AzureOpenAIClient(**client_kwargs)
        self._init = True

    async def __asetup__(self, context=None):
        if self._ainit:
            return

        client_kwargs = self._get_azure_client_params()
        self._async_client = AsyncAzureOpenAI(**client_kwargs)
        self._ainit = True

    def _get_azure_client_params(self) -> dict[str, Any]:
        """Build Azure-specific client parameters."""
        params: dict[str, Any] = {}

        if self.api_key is not None:
            params["api_key"] = self.api_key
        if self.api_version is not None:
            params["api_version"] = self.api_version
        if self.azure_endpoint is not None:
            params["azure_endpoint"] = self.azure_endpoint
        if self.azure_deployment is not None:
            params["azure_deployment"] = self.azure_deployment
        if self.azure_ad_token is not None:
            params["azure_ad_token"] = self.azure_ad_token
        if self.organization is not None:
            params["organization"] = self.organization
        if self.timeout is not None:
            params["timeout"] = self.timeout
        if self.max_retries is not None:
            params["max_retries"] = self.max_retries
        if self.default_headers is not None:
            params["default_headers"] = self.default_headers
        if self.default_query is not None:
            params["default_query"] = self.default_query
        if self.nparams:
            params.update(self.nparams)

        return params



import os
import time

import pytest

from x8.content.image import Image
from x8.content.video import Video
from x8.core.exceptions import BadRequestError

from ._providers import VideoGenerationProvider, get_judge
from ._sync_and_async_client import VideoGenerationSyncAndAsyncClient


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VideoGenerationProvider.OPENAI,
        VideoGenerationProvider.GOOGLE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_simple(provider_type: str, async_call: bool):
    client = VideoGenerationSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    result = await client.generate(prompt="dog and a cat")
    id = result.id
    while True:
        result = await client.get(id=id)
        if result.status in ["completed"]:
            break
        if result.status in ["failed"]:
            raise BadRequestError("Video generation failed")
        print(result.status)
        time.sleep(1)

    content = await client.download(id=id)
    chunks = (
        [chunk async for chunk in content] if async_call else list(content)
    )
    video = Video.load(b"".join(chunks))
    judge(video, keywords=["cat", "kitten", "dog"])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VideoGenerationProvider.OPENAI,
        VideoGenerationProvider.GOOGLE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_poll(provider_type: str, async_call: bool):
    client = VideoGenerationSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    result = await client.generate(prompt="dog and a cat playing", poll=True)
    video = Video.load(result.videos[0])
    judge(video, keywords=["cat", "kitten", "dog"])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VideoGenerationProvider.OPENAI,
        VideoGenerationProvider.GOOGLE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_stream(provider_type: str, async_call: bool):
    client = VideoGenerationSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    result_iterator = await client.generate(
        prompt="dog and a cat playing", stream=True
    )

    final_result = None
    if async_call:
        async for response in result_iterator:
            print(f"Status: {response.result.status}")
            final_result = response.result
            if response.result.status in ["completed", "failed"]:
                break
    else:
        for response in result_iterator:
            print(f"Status: {response.result.status}")
            final_result = response.result
            if response.result.status in ["completed", "failed"]:
                break

    if final_result.status == "failed":
        raise BadRequestError(f"Video generation failed: {final_result.error}")

    video = Video.load(final_result.videos[0])
    judge(video, keywords=["cat", "kitten", "dog"])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        VideoGenerationProvider.OPENAI,
        VideoGenerationProvider.GOOGLE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_image(provider_type: str, async_call: bool):
    client = VideoGenerationSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    # Load the dog image
    image_path = os.path.join(os.path.dirname(__file__), "images", "dog.jpg")
    image = Image.load(image_path)

    result = await client.generate(
        prompt="Make the dog run and jump", image=image.get_data(), poll=True
    )
    video = Video.load(result.videos[0])
    judge(video, keywords=["dog", "run", "jump"])


def judge(video, keywords: list[str]):
    frame = video.get_frame()

    judge = get_judge()
    result = judge.generate(
        input=[
            {
                "type": "message",
                "role": "user",
                "content": [
                    {
                        "type": "input_image",
                        "image": {
                            "content": frame.image.convert(
                                type="bytes", format="jpeg"
                            ),
                            "media_type": "image/jpeg",
                        },
                    },
                    {
                        "type": "input_text",
                        "text": "Describe this image in one sentence.",
                    },
                ],
            }
        ],
    )
    message = result.output[0]
    text_content = message.content[0].text.lower()
    print(text_content)
    for keyword in keywords:
        if keyword in text_content:
            return
    raise AssertionError(
        f"Keywords {keywords} not found in generated text: {text_content}"
    )



import base64
import uuid
from typing import Any, AsyncIterator, Iterator

import replicate
from replicate import Client as ReplicateClient
from replicate.stream import ServerSentEvent

from x8.core import Response
from x8.core._provider import Provider
from x8.core.exceptions import BadRequestError

from .._models import (
    InputItem,
    OutputItem,
    OutputMessage,
    OutputText,
    Reasoning,
    ResponseCompletedEvent,
    ResponseOutputTextDeltaEvent,
    ResponseText,
    TextGenerationResult,
    TextGenerationStreamEvent,
    Tool,
    ToolChoice,
    Usage,
)


class Replicate(Provider):
    """Replicate AI provider using the native Replicate SDK.

    Replicate provides access to many AI models including Gemini, GPT,
    Llama, and more through a unified API.

    Note: Replicate uses a prompt-based API rather than chat messages.
    Tool calling is not supported by most Replicate models.
    """

    api_key: str | None
    model: str
    max_tokens: int
    nparams: dict[str, Any] | None

    _client: ReplicateClient
    _init: bool
    _ainit: bool

    def __init__(
        self,
        api_key: str | None = None,
        model: str = "google/gemini-3-pro",
        max_tokens: int = 8192,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            api_key:
                Replicate API token.
            model:
                Replicate model to use for text generation.
                Popular models include:
                - google/gemini-3-pro
                - openai/gpt-4o
                - meta/llama-3.3-70b-instruct
            max_tokens:
                Default maximum tokens for responses.
            nparams:
                Native params for Replicate client.
        """
        self.api_key = api_key
        self.model = model
        self.max_tokens = max_tokens
        self.nparams = nparams
        self._init = False
        self._ainit = False
        super().__init__(**kwargs)

    def __setup__(self, context=None):
        if self._init:
            return
        client_kwargs = {}
        if self.api_key is not None:
            client_kwargs["api_token"] = self.api_key
        if self.nparams:
            client_kwargs.update(self.nparams)
        self._client = ReplicateClient(**client_kwargs)
        self._init = True

    async def __asetup__(self, context=None):
        if self._ainit:
            return
        # Replicate uses the same client for sync and async
        client_kwargs = {}
        if self.api_key is not None:
            client_kwargs["api_token"] = self.api_key
        if self.nparams:
            client_kwargs.update(self.nparams)
        self._client = ReplicateClient(**client_kwargs)
        self._ainit = True

    def generate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | Iterator[Response[TextGenerationStreamEvent]]
    ):
        model_ref, input_args = self._convert_args(
            input=input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            text=text,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        try:
            if not stream:
                output = self._client.run(model_ref, input=input_args)
                # Output is typically a list of strings or a single string
                if isinstance(output, list):
                    text_content = "".join(str(item) for item in output)
                else:
                    text_content = str(output) if output else ""

                result = self._convert_result(
                    text_content,
                    model=model_ref,
                )
                return Response(result=result)
            else:

                def _stream_iter() -> (
                    Iterator[Response[TextGenerationStreamEvent]]
                ):
                    accumulated_text = ""
                    for event in self._client.stream(
                        model_ref, input=input_args
                    ):
                        if isinstance(event, ServerSentEvent):
                            if event.event == ServerSentEvent.EventType.OUTPUT:
                                delta = event.data
                                accumulated_text += delta
                                yield Response(
                                    result=ResponseOutputTextDeltaEvent(
                                        sequence_number=None,
                                        content_index=0,
                                        output_index=0,
                                        delta=delta,
                                    )
                                )

                    # Emit completed event
                    final_result = self._build_final_result(
                        model=model_ref,
                        accumulated_text=accumulated_text,
                    )
                    yield Response(
                        result=ResponseCompletedEvent(
                            sequence_number=None, response=final_result
                        )
                    )

                return _stream_iter()
        except replicate.exceptions.ReplicateError as e:
            raise BadRequestError(str(e)) from e

    async def agenerate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | AsyncIterator[Response[TextGenerationStreamEvent]]
    ):
        model_ref, input_args = self._convert_args(
            input=input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            text=text,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        try:
            if not stream:
                output = await self._client.async_run(
                    model_ref, input=input_args
                )
                # Output is typically a list of strings or a single string
                if isinstance(output, list):
                    text_content = "".join(str(item) for item in output)
                else:
                    text_content = str(output) if output else ""

                result = self._convert_result(
                    text_content,
                    model=model_ref,
                )
                return Response(result=result)
            else:

                async def _poll_aiter() -> (
                    AsyncIterator[Response[TextGenerationStreamEvent]]
                ):
                    accumulated_text = ""
                    # async_stream returns a coroutine, must await first
                    stream = await self._client.async_stream(
                        model_ref, input=input_args
                    )
                    async for event in stream:
                        if isinstance(event, ServerSentEvent):
                            if event.event == ServerSentEvent.EventType.OUTPUT:
                                delta = event.data
                                accumulated_text += delta
                                yield Response(
                                    result=ResponseOutputTextDeltaEvent(
                                        sequence_number=None,
                                        content_index=0,
                                        output_index=0,
                                        delta=delta,
                                    )
                                )

                    # Emit completed event
                    final_result = self._build_final_result(
                        model=model_ref,
                        accumulated_text=accumulated_text,
                    )
                    yield Response(
                        result=ResponseCompletedEvent(
                            sequence_number=None, response=final_result
                        )
                    )

                return _poll_aiter()
        except replicate.exceptions.ReplicateError as e:
            raise BadRequestError(str(e)) from e

    def _convert_args(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[str, dict[str, Any]]:
        """Convert unified args to Replicate format."""
        # Model reference
        model_ref = model or self.model

        # Build input args for the model
        input_args: dict[str, Any] = {}

        # Convert input to prompt and extract images
        prompt, images = self._convert_input_to_prompt(input, instructions)
        input_args["prompt"] = prompt

        if images:
            input_args["images"] = images

        # System instruction (if model supports it)
        if instructions:
            input_args["system_instruction"] = instructions

        # Temperature
        if temperature is not None:
            input_args["temperature"] = temperature

        # Top P
        if top_p is not None:
            input_args["top_p"] = top_p

        # Max tokens
        if max_output_tokens is not None:
            input_args["max_output_tokens"] = max_output_tokens
        else:
            input_args["max_output_tokens"] = self.max_tokens

        # Reasoning / thinking level
        if reasoning is not None:
            if isinstance(reasoning, dict):
                reasoning = Reasoning(**reasoning)
            effort = reasoning.effort
            if effort == "none":
                pass  # No thinking
            elif effort == "low":
                input_args["thinking_level"] = "low"
            elif effort in ("medium", "high"):
                input_args["thinking_level"] = "high"

        # JSON schema mode - append instruction to prompt
        if text is not None:
            if isinstance(text, dict):
                text = ResponseText(**text)
            if text.format is not None:
                fmt = text.format
                if isinstance(fmt, dict):
                    fmt_type = fmt.get("type")
                    fmt_schema = fmt.get("schema", {})
                else:
                    fmt_type = getattr(fmt, "type", None)
                    fmt_schema = getattr(fmt, "schema_", {})

                if fmt_type in ("json_schema", "json_object"):
                    import json

                    schema_str = json.dumps(fmt_schema, indent=2)
                    input_args["prompt"] += (
                        f"\n\nRespond with valid JSON matching this schema:\n"
                        f"```json\n{schema_str}\n```"
                    )

        # Native config overrides
        if nconfig:
            input_args.update(nconfig)

        return model_ref, input_args

    def _convert_input_to_prompt(
        self,
        input: str | list[dict[str, Any] | InputItem],
        instructions: str | None = None,
    ) -> tuple[str, list[str]]:
        """Convert unified input format to prompt string and image list."""
        prompt_parts: list[str] = []
        images: list[str] = []

        # Handle simple string input
        if isinstance(input, str):
            return input, images

        # Handle list of items - build a conversation prompt
        for item in input:
            if isinstance(item, dict):
                item_type = item.get("type")
                if item_type == "message":
                    role = item.get("role", "user")
                    content = item.get("content", "")

                    if isinstance(content, str):
                        if role == "user":
                            prompt_parts.append(f"User: {content}")
                        elif role == "assistant":
                            prompt_parts.append(f"Assistant: {content}")
                        elif role == "system":
                            prompt_parts.append(f"System: {content}")
                    elif isinstance(content, list):
                        # Handle content parts
                        text_parts = []
                        for part in content:
                            if isinstance(part, dict):
                                part_type = part.get("type")
                                if part_type == "input_text":
                                    text_parts.append(part.get("text", ""))
                                elif part_type == "input_image":
                                    img = part.get("image", {})
                                    img_url = self._convert_image(img)
                                    if img_url:
                                        images.append(img_url)

                        if text_parts:
                            combined_text = " ".join(text_parts)
                            if role == "user":
                                prompt_parts.append(f"User: {combined_text}")
                            elif role == "assistant":
                                prompt_parts.append(
                                    f"Assistant: {combined_text}"
                                )
                            elif role == "system":
                                prompt_parts.append(f"System: {combined_text}")

                elif item_type == "function_call_output":
                    # Include tool results in the prompt
                    output = item.get("output", "")
                    prompt_parts.append(f"Tool Result: {output}")
            else:
                # Handle InputItem objects
                if hasattr(item, "type"):
                    if item.type == "message":
                        role = getattr(item, "role", "user")
                        content = getattr(item, "content", "")
                        if isinstance(content, str):
                            if role == "user":
                                prompt_parts.append(f"User: {content}")
                            elif role == "assistant":
                                prompt_parts.append(f"Assistant: {content}")

        # Join all parts
        if prompt_parts:
            return "\n\n".join(prompt_parts), images

        return "", images

    def _convert_image(self, image: dict[str, Any]) -> str | None:
        """Convert image to Replicate format (URL or data URI)."""
        if "url" in image:
            return image["url"]
        elif "content" in image:
            content = image["content"]
            media_type = image.get("media_type", "image/jpeg")
            if isinstance(content, bytes):
                b64 = base64.b64encode(content).decode("utf-8")
                return f"data:{media_type};base64,{b64}"
            elif isinstance(content, str):
                # Assume already base64 encoded
                return f"data:{media_type};base64,{content}"
        return None

    def _convert_result(
        self,
        text_content: str,
        model: str,
    ) -> TextGenerationResult:
        """Convert Replicate output to unified result format."""
        output_items: list[OutputItem] = []

        # Create message with text content
        message = OutputMessage(
            type="message",
            role="assistant",
            content=[OutputText(type="output_text", text=text_content)],
        )
        output_items.append(message)

        # Build usage (Replicate doesn't provide token counts in basic output)
        # Estimate tokens roughly: ~4 chars per token
        estimated_output_tokens = max(1, len(text_content) // 4)
        usage = Usage(
            input_tokens=1,  # Minimum to pass validation
            output_tokens=estimated_output_tokens,
            total_tokens=1 + estimated_output_tokens,
        )

        return TextGenerationResult(
            id=f"replicate-{uuid.uuid4().hex[:16]}",
            model=model,
            status="completed",
            output=output_items,
            usage=usage,
        )

    def _build_final_result(
        self,
        model: str,
        accumulated_text: str,
    ) -> TextGenerationResult:
        """Build final result from accumulated stream data."""
        output_items: list[OutputItem] = []

        # Create message with accumulated text
        if accumulated_text:
            message = OutputMessage(
                type="message",
                role="assistant",
                content=[
                    OutputText(type="output_text", text=accumulated_text)
                ],
            )
            output_items.append(message)

        # Estimate tokens roughly: ~4 chars per token
        estimated_output_tokens = max(1, len(accumulated_text) // 4)
        usage = Usage(
            input_tokens=1,  # Minimum to pass validation
            output_tokens=estimated_output_tokens,
            total_tokens=1 + estimated_output_tokens,
        )

        return TextGenerationResult(
            id=f"replicate-{uuid.uuid4().hex[:16]}",
            model=model,
            status="completed",
            output=output_items,
            usage=usage,
        )



import base64
import json
from typing import Any, AsyncIterator, Iterator, Literal, Mapping

import anthropic

from x8.core import Response
from x8.core._provider import Provider
from x8.core.exceptions import BadRequestError

from .._models import (
    AllowedTools,
    ErrorDetail,
    FunctionCall,
    InputItem,
    OutputItem,
    OutputMessage,
    OutputMessageContent,
    OutputReasoning,
    OutputReasoningContentText,
    OutputText,
    Reasoning,
    ResponseCompletedEvent,
    ResponseFunctionCallArgumentsDeltaEvent,
    ResponseOutputTextDeltaEvent,
    ResponseReasoningTextDeltaEvent,
    ResponseText,
    TextGenerationResult,
    TextGenerationStreamEvent,
    Tool,
    ToolChoice,
    ToolChoiceFunction,
    Usage,
)


class Anthropic(Provider):
    api_key: str | None
    auth_token: str | None
    base_url: str | None
    timeout: float | None
    max_retries: int | None
    default_headers: Mapping[str, str] | None
    default_query: Mapping[str, object] | None
    model: str
    max_tokens: int
    nparams: dict[str, Any] | None

    _client: anthropic.Anthropic
    _async_client: anthropic.AsyncAnthropic
    _init: bool
    _ainit: bool

    def __init__(
        self,
        api_key: str | None = None,
        auth_token: str | None = None,
        base_url: str | None = None,
        timeout: float | None = None,
        max_retries: int | None = None,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        model: str = "claude-sonnet-4-20250514",
        max_tokens: int = 16384,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            api_key:
                Anthropic API key.
            auth_token:
                Anthropic auth token (alternative to api_key).
            base_url:
                Anthropic base url for API requests.
            timeout:
                Timeout for client requests.
            max_retries:
                Maximum number of retries for failed requests.
            default_headers:
                Default headers to include in every request.
            default_query:
                Default query parameters to include in every request.
            model:
                Anthropic model to use for text generation.
            max_tokens:
                Default maximum tokens for responses.
            nparams:
                Native params for Anthropic client.
        """
        self.api_key = api_key
        self.auth_token = auth_token
        self.base_url = base_url
        self.timeout = timeout
        self.max_retries = max_retries
        self.default_headers = default_headers
        self.default_query = default_query
        self.model = model
        self.max_tokens = max_tokens
        self.nparams = nparams
        self._init = False
        self._ainit = False
        super().__init__(**kwargs)

    def __setup__(self, context=None):
        if self._init:
            return
        client_kwargs = {}
        if self.api_key is not None:
            client_kwargs["api_key"] = self.api_key
        if self.auth_token is not None:
            client_kwargs["auth_token"] = self.auth_token
        if self.base_url is not None:
            client_kwargs["base_url"] = self.base_url
        if self.timeout is not None:
            client_kwargs["timeout"] = self.timeout
        if self.max_retries is not None:
            client_kwargs["max_retries"] = self.max_retries
        if self.default_headers is not None:
            client_kwargs["default_headers"] = self.default_headers
        if self.default_query is not None:
            client_kwargs["default_query"] = self.default_query
        if self.nparams:
            client_kwargs.update(self.nparams)
        self._client = anthropic.Anthropic(**client_kwargs)
        self._init = True

    async def __asetup__(self, context=None):
        if self._ainit:
            return
        client_kwargs = {}
        if self.api_key is not None:
            client_kwargs["api_key"] = self.api_key
        if self.auth_token is not None:
            client_kwargs["auth_token"] = self.auth_token
        if self.base_url is not None:
            client_kwargs["base_url"] = self.base_url
        if self.timeout is not None:
            client_kwargs["timeout"] = self.timeout
        if self.max_retries is not None:
            client_kwargs["max_retries"] = self.max_retries
        if self.default_headers is not None:
            client_kwargs["default_headers"] = self.default_headers
        if self.default_query is not None:
            client_kwargs["default_query"] = self.default_query
        if self.nparams:
            client_kwargs.update(self.nparams)
        self._async_client = anthropic.AsyncAnthropic(**client_kwargs)
        self._ainit = True

    def generate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | Iterator[Response[TextGenerationStreamEvent]]
    ):
        args = self._convert_generate_args(
            input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            text=text,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        # Check if we're using JSON schema mode (tool-based)
        json_schema_mode = args.pop("_json_schema_mode", False)
        json_schema_tool_name = args.pop("_json_schema_tool_name", None)
        try:
            if not stream:
                response = self._client.messages.create(**args)
                result = self._convert_result(
                    response,
                    json_schema_mode=json_schema_mode,
                    json_schema_tool_name=json_schema_tool_name,
                )
                return Response(result=result)
            else:

                def _stream_iter() -> (
                    Iterator[Response[TextGenerationStreamEvent]]
                ):
                    with self._client.messages.stream(**args) as stream_ctx:
                        for event in stream_ctx:
                            converted_events = self._convert_stream_event(
                                event,
                                json_schema_mode=json_schema_mode,
                            )
                            for converted_event in converted_events:
                                yield Response(result=converted_event)
                        # Final completed event with the full message
                        final_message = stream_ctx.get_final_message()
                        if final_message:
                            yield Response(
                                result=ResponseCompletedEvent(
                                    response=self._convert_result(
                                        final_message,
                                        json_schema_mode=json_schema_mode,
                                        json_schema_tool_name=json_schema_tool_name,  # noqa
                                    )
                                )
                            )

                return _stream_iter()
        except anthropic.BadRequestError as e:
            raise BadRequestError(str(e.message)) from e

    async def agenerate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | AsyncIterator[Response[TextGenerationStreamEvent]]
    ):
        args = self._convert_generate_args(
            input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            text=text,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        # Check if we're using JSON schema mode (tool-based)
        json_schema_mode = args.pop("_json_schema_mode", False)
        json_schema_tool_name = args.pop("_json_schema_tool_name", None)
        try:
            if not stream:
                response = await self._async_client.messages.create(**args)
                result = self._convert_result(
                    response,
                    json_schema_mode=json_schema_mode,
                    json_schema_tool_name=json_schema_tool_name,
                )
                return Response(result=result)
            else:

                async def _poll_aiter() -> (
                    AsyncIterator[Response[TextGenerationStreamEvent]]
                ):
                    async with self._async_client.messages.stream(
                        **args
                    ) as stream_ctx:
                        async for event in stream_ctx:
                            converted_events = self._convert_stream_event(
                                event,
                                json_schema_mode=json_schema_mode,
                            )
                            for converted_event in converted_events:
                                yield Response(result=converted_event)
                        # Final completed event with the full message
                        final_message = await stream_ctx.get_final_message()
                        if final_message:
                            yield Response(
                                result=ResponseCompletedEvent(
                                    response=self._convert_result(
                                        final_message,
                                        json_schema_mode=json_schema_mode,
                                        json_schema_tool_name=json_schema_tool_name,  # noqa
                                    )
                                )
                            )

                return _poll_aiter()
        except anthropic.BadRequestError as e:
            raise BadRequestError(str(e.message)) from e

    def _convert_stream_event(
        self,
        event: Any,
        json_schema_mode: bool = False,
    ) -> list[TextGenerationStreamEvent]:
        """Convert Anthropic streaming events to unified format."""
        results: list[TextGenerationStreamEvent] = []

        event_type = getattr(event, "type", None)

        # Text delta events
        if event_type == "content_block_delta":
            delta = getattr(event, "delta", None)
            if delta:
                delta_type = getattr(delta, "type", None)
                if delta_type == "text_delta":
                    text = getattr(delta, "text", "")
                    if text:
                        results.append(
                            ResponseOutputTextDeltaEvent(delta=text)
                        )
                elif delta_type == "thinking_delta":
                    thinking = getattr(delta, "thinking", "")
                    if thinking:
                        results.append(
                            ResponseReasoningTextDeltaEvent(delta=thinking)
                        )
                elif delta_type == "input_json_delta":
                    # Tool call arguments delta
                    partial_json = getattr(delta, "partial_json", "")
                    if partial_json:
                        # In JSON schema mode, emit as text delta
                        if json_schema_mode:
                            results.append(
                                ResponseOutputTextDeltaEvent(
                                    delta=partial_json
                                )
                            )
                        else:
                            results.append(
                                ResponseFunctionCallArgumentsDeltaEvent(
                                    delta=partial_json
                                )
                            )

        # Message complete event - handled via get_final_message instead
        # to get the complete message with usage info

        return results

    def _convert_result(
        self,
        response: Any,
        json_schema_mode: bool = False,
        json_schema_tool_name: str | None = None,
    ) -> TextGenerationResult:
        """Convert Anthropic response to unified TextGenerationResult."""
        r: dict[str, Any] = (
            response.model_dump()
            if hasattr(response, "model_dump")
            else response
        )

        output_items: list[OutputItem] = []
        message_content: list[OutputMessageContent] = []
        reasoning_content: list[OutputReasoningContentText] = []
        function_calls: list[FunctionCall] = []

        # Process content blocks
        content = r.get("content", [])
        if isinstance(content, list):
            for block in content:
                if not isinstance(block, dict):
                    continue

                block_type = block.get("type")

                if block_type == "text":
                    text = block.get("text", "")
                    if text:
                        message_content.append(OutputText(text=text))

                elif block_type == "thinking":
                    thinking_text = block.get("thinking", "")
                    if thinking_text:
                        reasoning_content.append(
                            OutputReasoningContentText(text=thinking_text)
                        )

                elif block_type == "tool_use":
                    tool_name = block.get("name", "")
                    tool_input = block.get("input")

                    # Check if this is our JSON schema tool
                    if (
                        json_schema_mode
                        and json_schema_tool_name
                        and tool_name == json_schema_tool_name
                    ):
                        # Convert tool output to text message with JSON
                        json_text = json.dumps(tool_input)
                        message_content.append(OutputText(text=json_text))
                    else:
                        function_calls.append(
                            FunctionCall(
                                name=tool_name,
                                arguments=tool_input,
                                call_id=block.get("id"),
                                id=block.get("id"),
                            )
                        )

        # Add reasoning as separate output item if present
        if reasoning_content:
            output_items.append(
                OutputReasoning(
                    content=reasoning_content,
                    status="completed",
                )
            )

        # Add assistant message if any text content collected
        if message_content:
            output_items.append(
                OutputMessage(
                    role="assistant",
                    content=message_content,
                    status="completed",
                )
            )

        # Add function calls as separate output items
        for fc in function_calls:
            output_items.append(fc)

        # Usage mapping
        usage_obj: Usage | None = None
        usage = r.get("usage")
        if isinstance(usage, dict):
            input_tokens = int(usage.get("input_tokens") or 0)
            output_tokens = int(usage.get("output_tokens") or 0)
            cache_read = int(usage.get("cache_read_input_tokens") or 0)
            cache_creation = int(usage.get("cache_creation_input_tokens") or 0)
            total_tokens = input_tokens + output_tokens

            input_details: dict[str, int] | None = None
            if cache_read or cache_creation:
                input_details = {}
                if cache_read:
                    input_details["cache_read_tokens"] = cache_read
                if cache_creation:
                    input_details["cache_creation_tokens"] = cache_creation

            usage_obj = Usage(
                input_tokens=input_tokens,
                input_tokens_details=input_details,
                output_tokens=output_tokens,
                total_tokens=total_tokens,
            )

        # Error handling
        error_obj: ErrorDetail | None = None

        # Determine status based on stop_reason
        status: Literal[
            "completed",
            "failed",
            "in_progress",
            "cancelled",
            "queued",
            "incomplete",
        ] = "completed"
        stop_reason = r.get("stop_reason")
        if stop_reason == "max_tokens":
            status = "incomplete"
        elif stop_reason == "end_turn":
            status = "completed"
        elif stop_reason == "tool_use":
            status = "completed"
        elif stop_reason == "stop_sequence":
            status = "completed"

        result = TextGenerationResult(
            id=r.get("id"),
            model=r.get("model"),
            created_at=None,  # Anthropic doesn't provide creation time
            status=status,
            error=error_obj,
            output=output_items or None,
            usage=usage_obj,
        )
        return result

    def _convert_generate_args(
        self,
        input: str | list[dict[str, Any] | InputItem],
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        args: dict[str, Any] = {}
        args["model"] = model or self.model
        args["max_tokens"] = max_output_tokens or self.max_tokens

        messages: list[dict[str, Any]] = []

        def _convert_content_part(c: dict[str, Any]) -> dict[str, Any] | None:
            """Convert a single content part to Anthropic format."""
            ct = c.get("type")
            if ct == "input_text":
                return {"type": "text", "text": c.get("text")}
            if ct == "output_text":
                return {"type": "text", "text": c.get("text")}
            if ct == "input_image":
                img = c.get("image")
                if not isinstance(img, dict):
                    return None
                media_type = img.get("media_type") or "image/jpeg"
                content = img.get("content")
                source = img.get("source")
                if isinstance(content, (bytes, bytearray)):
                    b64_data = base64.b64encode(content).decode("utf-8")
                    return {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": media_type,
                            "data": b64_data,
                        },
                    }
                elif isinstance(content, str):
                    # Already base64 encoded
                    return {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": media_type,
                            "data": content,
                        },
                    }
                if isinstance(source, str):
                    # URL source
                    return {
                        "type": "image",
                        "source": {
                            "type": "url",
                            "url": source,
                        },
                    }
            if ct == "input_file":
                f = c.get("file")
                if isinstance(f, dict):
                    content = f.get("content")
                    mime_type = f.get("media_type") or "application/pdf"
                    if isinstance(content, (bytes, bytearray)):
                        b64_data = base64.b64encode(content).decode("utf-8")
                        return {
                            "type": "document",
                            "source": {
                                "type": "base64",
                                "media_type": mime_type,
                                "data": b64_data,
                            },
                        }
                    elif isinstance(content, str):
                        # Already base64 encoded
                        return {
                            "type": "document",
                            "source": {
                                "type": "base64",
                                "media_type": mime_type,
                                "data": content,
                            },
                        }
                    source = f.get("source")
                    if isinstance(source, str):
                        return {
                            "type": "document",
                            "source": {
                                "type": "url",
                                "url": source,
                            },
                        }
            return None

        def _convert_content_to_parts(
            raw_content: str | list | None,
        ) -> list[dict[str, Any]]:
            """Convert message content to Anthropic format."""
            if isinstance(raw_content, str):
                return [{"type": "text", "text": raw_content}]
            if isinstance(raw_content, list):
                parts = []
                for c in raw_content:
                    if isinstance(c, dict):
                        part = _convert_content_part(c)
                        if part:
                            parts.append(part)
                return parts
            return []

        def _convert_input_item(msg: InputItem | dict[str, Any]) -> None:
            m: dict[str, Any] = msg if isinstance(msg, dict) else msg.to_dict()
            item_type = m.get("type")

            if item_type == "function_call":
                # Anthropic represents function calls as tool_use blocks
                # in assistant messages
                tool_use_block = {
                    "type": "tool_use",
                    "id": m.get("call_id") or m.get("id"),
                    "name": m.get("name"),
                    "input": m.get("arguments") or {},
                }
                messages.append(
                    {
                        "role": "assistant",
                        "content": [tool_use_block],
                    }
                )
                return

            if item_type == "function_call_output":
                # Tool results go in user messages with tool_result blocks
                output = m.get("output")
                if isinstance(output, dict):
                    output = json.dumps(output)
                messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "tool_result",
                                "tool_use_id": m.get("call_id"),
                                "content": str(output) if output else "",
                            }
                        ],
                    }
                )
                return

            # Handle OutputReasoning (skip)
            if item_type == "reasoning":
                return

            if item_type != "message":
                return

            role = m.get("role")
            # Map roles for Anthropic
            if role == "system" or role == "developer":
                # System messages handled via system parameter
                content = m.get("content")
                if isinstance(content, str):
                    if "system" not in args:
                        args["system"] = content
                    else:
                        args["system"] += "\n" + content
                return
            elif role == "assistant":
                a_role = "assistant"
            else:
                a_role = "user"

            parts = _convert_content_to_parts(m.get("content"))
            if parts:
                messages.append({"role": a_role, "content": parts})

        # Input may be a raw string or a sequence of InputItems.
        if isinstance(input, str):
            messages.append(
                {"role": "user", "content": [{"type": "text", "text": input}]}
            )
        else:
            for it in input:
                _convert_input_item(it)

        args["messages"] = messages

        if temperature is not None:
            args["temperature"] = temperature
        if top_p is not None:
            args["top_p"] = top_p
        if instructions:
            args["system"] = instructions

        # Handle JSON output format via tool-based approach
        json_schema_tool_name: str | None = None
        if text:
            text_dict = text if isinstance(text, dict) else text.to_dict()
            fmt = text_dict.get("format")
            if isinstance(fmt, dict) and fmt.get("type") == "json_schema":
                # Create a tool with the JSON schema to force structured output
                json_schema_tool_name = fmt.get("name", "json_output")
                schema = fmt.get(
                    "schema", {"type": "object", "properties": {}}
                )
                description = fmt.get(
                    "description", "Generate structured JSON output"
                )

                # Create the tool
                json_tool = {
                    "name": json_schema_tool_name,
                    "description": description,
                    "input_schema": schema,
                }

                # Add to existing tools or create new tools list
                if "tools" not in args:
                    args["tools"] = []
                args["tools"].append(json_tool)

                # Force the model to use this tool
                args["tool_choice"] = {
                    "type": "tool",
                    "name": json_schema_tool_name,
                }

                # Set markers for result conversion
                args["_json_schema_mode"] = True
                args["_json_schema_tool_name"] = json_schema_tool_name

        # Handle tools
        if tools:
            args["tools"] = self._convert_tools(tools)
        if tool_choice:
            args["tool_choice"] = self._convert_tool_choice(tool_choice)

        # Handle reasoning/thinking
        if reasoning:
            if isinstance(reasoning, dict):
                effort = reasoning.get("effort", "none")
            elif isinstance(reasoning, Reasoning):
                effort = reasoning.effort
            else:
                effort = "none"

            if effort and effort != "none":
                # Enable extended thinking
                budget_map = {
                    "low": 5000,
                    "medium": 10000,
                    "high": 20000,
                }
                budget = budget_map.get(effort, 5000)
                args["thinking"] = {
                    "type": "enabled",
                    "budget_tokens": budget,
                }

        if nconfig:
            args.update(nconfig)

        return args

    def _convert_tools(self, tools: list[dict | Tool]) -> list[dict[str, Any]]:
        """Convert tools to Anthropic format."""
        anthropic_tools: list[dict[str, Any]] = []
        for t in tools:
            if isinstance(t, dict):
                tool_type = t.get("type", "function")
                tool = t
            else:
                tool_type = t.type
                tool = t.to_dict()

            if tool_type == "function":
                anthropic_tools.append(
                    {
                        "name": tool.get("name"),
                        "description": tool.get("description"),
                        "input_schema": tool.get("parameters")
                        or {
                            "type": "object",
                            "properties": {},
                        },
                    }
                )
            elif tool_type == "web_search":
                # Anthropic's web search tool
                web_search_tool: dict[str, Any] = {
                    "type": "web_search_20250305",
                    "name": "web_search",
                }
                # Optional: map search_context_size to max_uses
                search_context_size = tool.get("search_context_size", "medium")
                max_uses_map = {
                    "short": 1,
                    "medium": 3,
                    "long": 5,
                }
                if search_context_size in max_uses_map:
                    web_search_tool["max_uses"] = max_uses_map[
                        search_context_size
                    ]
                # Optional: allowed_domains filter
                filters = tool.get("filters")
                if isinstance(filters, dict):
                    allowed_domains = filters.get("allowed_domains")
                    if allowed_domains:
                        web_search_tool["allowed_domains"] = allowed_domains
                # Optional: user_location
                user_location = tool.get("user_location")
                if isinstance(user_location, dict):
                    web_search_tool["user_location"] = {
                        "type": "approximate",
                        "city": user_location.get("city"),
                        "region": user_location.get("region"),
                        "country": user_location.get("country"),
                        "timezone": user_location.get("timezone"),
                    }
                anthropic_tools.append(web_search_tool)
            else:
                raise BadRequestError(f"Unsupported tool type: {tool_type}")

        return anthropic_tools

    def _convert_tool_choice(
        self,
        choice: dict | ToolChoice | None,
    ) -> dict[str, Any] | None:
        """Convert tool_choice to Anthropic format."""
        if choice is None:
            return None

        if isinstance(choice, (ToolChoiceFunction, AllowedTools)):
            inp = choice.to_dict()
        else:
            inp = choice

        if inp == "none":
            # Anthropic doesn't have a "none" mode - omit tools instead
            return None
        if inp == "auto":
            return {"type": "auto"}
        if inp == "required":
            return {"type": "any"}

        if isinstance(inp, dict):
            if inp.get("type") == "function":
                name = inp.get("name")
                if not name:
                    raise BadRequestError(
                        "Function tool_choice must specify a name"
                    )
                return {"type": "tool", "name": name}
            elif inp.get("type") == "allowed_tools":
                mode = inp.get("mode")
                if mode == "none":
                    return None
                if mode == "auto":
                    return {"type": "auto"}
                if mode == "required":
                    return {"type": "any"}

        raise BadRequestError(f"Unsupported tool_choice: {type(choice)}")



from ._models import (
    AllowedTools,
    ErrorDetail,
    ErrorEvent,
    Function,
    FunctionCall,
    FunctionCallOutput,
    InputFile,
    InputImage,
    InputItem,
    InputMessage,
    InputMessageContent,
    InputText,
    OutputItem,
    OutputMessage,
    OutputMessageContent,
    OutputReasoning,
    OutputReasoningContentText,
    OutputReasoningSummaryText,
    OutputText,
    Reasoning,
    Refusal,
    ResponseCompletedEvent,
    ResponseContentPartAddedEvent,
    ResponseContentPartDoneEvent,
    ResponseCreatedEvent,
    ResponseFailedEvent,
    ResponseFormatJSONSchema,
    ResponseFormatText,
    ResponseFunctionCallArgumentsDeltaEvent,
    ResponseFunctionCallArgumentsDoneEvent,
    ResponseIncompleteEvent,
    ResponseInProgressEvent,
    ResponseOutputItemAddedEvent,
    ResponseOutputItemDoneEvent,
    ResponseOutputTextDeltaEvent,
    ResponseOutputTextDoneEvent,
    ResponseQueuedEvent,
    ResponseReasoningSummaryPartAddedEvent,
    ResponseReasoningSummaryPartDoneEvent,
    ResponseReasoningSummaryTextDeltaEvent,
    ResponseReasoningSummaryTextDoneEvent,
    ResponseReasoningTextDeltaEvent,
    ResponseReasoningTextDoneEvent,
    ResponseRefusalDeltaEvent,
    ResponseRefusalDoneEvent,
    ResponseText,
    StreamEvent,
    TextGenerationResult,
    TextGenerationStreamEvent,
    Tool,
    ToolChoice,
    ToolChoiceFunction,
    Usage,
    WebSearch,
    WebSearchFilters,
    WebSearchUserLocation,
)
from .component import TextGeneration

__all__ = [
    "AllowedTools",
    "ErrorDetail",
    "ErrorEvent",
    "Function",
    "FunctionCall",
    "FunctionCallOutput",
    "InputFile",
    "InputImage",
    "InputItem",
    "InputMessage",
    "InputMessageContent",
    "InputText",
    "OutputItem",
    "OutputMessage",
    "OutputMessageContent",
    "OutputReasoning",
    "OutputReasoningContentText",
    "OutputReasoningSummaryText",
    "OutputText",
    "Reasoning",
    "Refusal",
    "ResponseCompletedEvent",
    "ResponseContentPartAddedEvent",
    "ResponseContentPartDoneEvent",
    "ResponseCreatedEvent",
    "ResponseFailedEvent",
    "ResponseFunctionCallArgumentsDeltaEvent",
    "ResponseFunctionCallArgumentsDoneEvent",
    "ResponseIncompleteEvent",
    "ResponseInProgressEvent",
    "ResponseOutputItemAddedEvent",
    "ResponseOutputItemDoneEvent",
    "ResponseOutputTextDeltaEvent",
    "ResponseOutputTextDoneEvent",
    "ResponseQueuedEvent",
    "ResponseReasoningSummaryPartAddedEvent",
    "ResponseReasoningSummaryPartDoneEvent",
    "ResponseReasoningSummaryTextDeltaEvent",
    "ResponseReasoningSummaryTextDoneEvent",
    "ResponseReasoningTextDeltaEvent",
    "ResponseReasoningTextDoneEvent",
    "ResponseRefusalDeltaEvent",
    "ResponseRefusalDoneEvent",
    "StreamEvent",
    "TextGenerationResult",
    "TextGenerationStreamEvent",
    "Tool",
    "ToolChoice",
    "ToolChoiceFunction",
    "Usage",
    "WebSearch",
    "WebSearchFilters",
    "WebSearchUserLocation",
    "TextGeneration",
    "ResponseFormatJSONSchema",
    "ResponseFormatText",
    "ResponseText",
]



import pytest

from x8.communication.email_service import EmailSendResult
from x8.core.exceptions import BadRequestError

from ._providers import EmailServiceProvider
from ._sync_and_async_client import EmailServiceSyncAndAsyncClient

TEST_FROM = "twoamowl@gmail.com"
TEST_TO = "slenin@gmail.com"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False],
)
async def test_send_simple_text(provider_type: str, async_call: bool):
    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    result = await client.send(
        to=TEST_TO,
        content="Hello from x8 test",
        from_email=TEST_FROM,
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.status is not None
    assert result.channel == "email"
    assert result.provider == "sendgrid"
    assert result.from_ == TEST_FROM
    assert TEST_TO in result.to


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_html(provider_type: str, async_call: bool):
    from x8.communication.email_service import EmailContent

    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    content = EmailContent(
        subject="HTML Test",
        html="<h1>Hello</h1><p>From x8 test</p>",
    )

    result = await client.send(
        to=TEST_TO,
        content=content,
        from_email=TEST_FROM,
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.status is not None
    assert result.provider == "sendgrid"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_with_subject(provider_type: str, async_call: bool):
    from x8.communication.email_service import EmailContent

    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    content = EmailContent(
        subject="Test Subject from x8",
        text="Email body with custom subject",
    )

    result = await client.send(
        to=TEST_TO,
        content=content,
        from_email=TEST_FROM,
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.provider == "sendgrid"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_multiple_recipients(provider_type: str, async_call: bool):
    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    recipients = ["recipient1@example.com", "recipient2@example.com"]

    result = await client.send(
        to=recipients,
        content="Hello to multiple recipients",
        from_email=TEST_FROM,
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert len(result.to) == 2
    assert result.to == recipients


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_with_cc_and_bcc(provider_type: str, async_call: bool):
    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    result = await client.send(
        to=TEST_TO,
        content="Email with CC and BCC",
        from_email=TEST_FROM,
        cc="cc@example.com",
        bcc="bcc@example.com",
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.provider == "sendgrid"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_with_reply_to(provider_type: str, async_call: bool):
    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    result = await client.send(
        to=TEST_TO,
        content="Email with reply-to",
        from_email=TEST_FROM,
        reply_to="replyto@example.com",
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.provider == "sendgrid"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_with_attachment(provider_type: str, async_call: bool):
    from x8.communication.email_service import EmailAttachment, EmailContent

    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    content = EmailContent(
        subject="Email with Attachment",
        text="Please see the attached file.",
    )

    attachment = EmailAttachment(
        filename="test.txt",
        content="This is test file content",
        content_type="text/plain",
    )

    result = await client.send(
        to=TEST_TO,
        content=content,
        from_email=TEST_FROM,
        attachments=[attachment],
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.provider == "sendgrid"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_with_text_and_html(provider_type: str, async_call: bool):
    from x8.communication.email_service import EmailContent

    client = EmailServiceSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    content = EmailContent(
        subject="Multipart Email",
        text="Plain text version",
        html="<h1>HTML version</h1>",
    )

    result = await client.send(
        to=TEST_TO,
        content=content,
        from_email=TEST_FROM,
    )

    assert isinstance(result, EmailSendResult)
    assert result.id is not None
    assert result.provider == "sendgrid"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_missing_sender_raises_error(
    provider_type: str, async_call: bool
):
    from x8.communication.email_service import EmailService

    # Create component without from_email
    component = EmailService(
        __unpack__=True,
        __provider__=dict(
            type=provider_type,
            parameters={"api_key": "test-key"},
        ),
    )

    with pytest.raises(BadRequestError):
        if async_call:
            await component.asend(to=TEST_TO, content="No sender")
        else:
            component.send(to=TEST_TO, content="No sender")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        EmailServiceProvider.SENDGRID,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_send_missing_content_raises_error(
    provider_type: str, async_call: bool
):
    from x8.communication.email_service import EmailContent, EmailService

    component = EmailService(
        __unpack__=True,
        __provider__=dict(
            type=provider_type,
            parameters={"api_key": "test-key"},
        ),
    )

    content = EmailContent(subject="No body")

    with pytest.raises(BadRequestError):
        if async_call:
            await component.asend(
                to=TEST_TO, content=content, from_email=TEST_FROM
            )
        else:
            component.send(to=TEST_TO, content=content, from_email=TEST_FROM)



from ._models import (
    MessageChannel,
    MessageContent,
    MessageMedia,
    MessageSendResult,
    VerificationCheckResult,
    VerificationStartResult,
)
from .component import MessageService

__all__ = [
    "MessageService",
    "MessageChannel",
    "MessageContent",
    "MessageMedia",
    "MessageSendResult",
    "VerificationStartResult",
    "VerificationCheckResult",
]



import base64
import json
from typing import Any, AsyncIterator, Iterator, Literal

from google.genai import Client
from google.genai import errors as google_errors
from google.genai import types

from x8._common.google_provider import GoogleProvider
from x8.core import Response
from x8.core.exceptions import BadRequestError

from .._models import (
    AllowedTools,
    ErrorDetail,
    FunctionCall,
    InputItem,
    OutputItem,
    OutputMessage,
    OutputMessageContent,
    OutputReasoning,
    OutputReasoningContentText,
    OutputText,
    Reasoning,
    ResponseCompletedEvent,
    ResponseFailedEvent,
    ResponseFunctionCallArgumentsDeltaEvent,
    ResponseIncompleteEvent,
    ResponseOutputTextDeltaEvent,
    ResponseReasoningTextDeltaEvent,
    ResponseText,
    TextGenerationResult,
    TextGenerationStreamEvent,
    Tool,
    ToolChoice,
    ToolChoiceFunction,
    Usage,
)


class Google(GoogleProvider):
    vertexai: bool
    project: str | None
    location: str
    model: str
    api_key: str | None
    nparams: dict[str, Any] | None

    _client: Client
    _init: bool

    def __init__(
        self,
        vertexai: bool = True,
        project: str | None = None,
        location: str = "global",
        model: str = "gemini-3-flash-preview",
        api_key: str | None = None,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        self.vertexai = vertexai
        self.project = project
        self.location = location
        self.model = model
        self.api_key = api_key
        self.nparams = nparams
        self._init = False
        super().__init__(
            service_account_info=service_account_info,
            service_account_file=service_account_file,
            access_token=access_token,
            **kwargs,
        )

    def __setup__(self, context=None):
        if self._init:
            return
        if self.api_key:
            credentials = None
        else:
            credentials = self._get_credentials()
        self._client = Client(
            vertexai=self.vertexai,
            api_key=self.api_key,
            credentials=credentials,
            project=self._get_project_or_default(self.project),
            location=self.location,
            **self.nparams or {},
        )
        self._init = True

    async def __asetup__(self, context=None):
        return self.__setup__(context)

    def generate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | Iterator[Response[TextGenerationStreamEvent]]
    ):
        args = self._convert_generate_args(
            input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            text=text,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        try:
            if not stream:
                response = self._client.models.generate_content(**args)
                result = self._convert_result(response)
                return Response(result=result)
            else:

                def _stream_iter() -> (
                    Iterator[Response[TextGenerationStreamEvent]]
                ):
                    response = self._client.models.generate_content_stream(
                        **args
                    )
                    for event in response:
                        converted_events = self._convert_stream_event(event)
                        for converted_event in converted_events:
                            yield Response(result=converted_event)

                return _stream_iter()
        except google_errors.ClientError as e:
            raise BadRequestError(str(e)) from e

    async def agenerate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | AsyncIterator[Response[TextGenerationStreamEvent]]
    ):
        args = self._convert_generate_args(
            input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            text=text,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        try:
            if not stream:
                response = await self._client.aio.models.generate_content(
                    **args
                )
                result = self._convert_result(response)
                return Response(result=result)
            else:

                async def _poll_aiter() -> (
                    AsyncIterator[Response[TextGenerationStreamEvent]]
                ):
                    response = (
                        await self._client.aio.models.generate_content_stream(
                            **args
                        )
                    )
                    async for event in response:
                        converted_events = self._convert_stream_event(event)
                        for converted_event in converted_events:
                            yield Response(result=converted_event)

                return _poll_aiter()
        except google_errors.ClientError as e:
            raise BadRequestError(str(e)) from e

    def _convert_stream_event(
        self, event: Any
    ) -> list[TextGenerationStreamEvent]:
        # Google streaming returns GenerateContentResponse chunks.
        # We synthesize unified stream events based on available data.
        # Returns a list because a single chunk
        # may contain both content and finish_reason.
        results: list[TextGenerationStreamEvent] = []

        try:
            e: dict[str, Any] = (
                event.model_dump() if hasattr(event, "model_dump") else event
            )
        except Exception:
            e = {}

        # Check candidates for content and finish state
        candidates = e.get("candidates")
        candidate0: dict[str, Any] | None = None
        if isinstance(candidates, list) and candidates:
            c0 = candidates[0]
            if isinstance(c0, dict):
                candidate0 = c0

        # Process parts for delta events FIRST (before terminal events)
        if candidate0:
            content = candidate0.get("content")
            if isinstance(content, dict):
                parts = content.get("parts")
                if isinstance(parts, list):
                    for part in parts:
                        if not isinstance(part, dict):
                            continue

                        # Text delta (reasoning or regular)
                        txt = part.get("text")
                        if isinstance(txt, str) and txt:
                            if part.get("thought"):
                                results.append(
                                    ResponseReasoningTextDeltaEvent(delta=txt)
                                )
                            else:
                                results.append(
                                    ResponseOutputTextDeltaEvent(delta=txt)
                                )

                        # Function call arguments delta
                        fc = part.get("function_call")
                        if isinstance(fc, dict):
                            args = fc.get("args")
                            if args is not None:
                                args_str = (
                                    json.dumps(args)
                                    if isinstance(args, dict)
                                    else str(args)
                                )
                                results.append(
                                    ResponseFunctionCallArgumentsDeltaEvent(
                                        delta=args_str,
                                        item_id=fc.get("id"),
                                    )
                                )

        # Fallback: try to get text directly from event if no parts found
        if not results:
            try:
                text = getattr(event, "text", None)
            except Exception:
                text = None

            if isinstance(text, str) and text:
                results.append(ResponseOutputTextDeltaEvent(delta=text))

        # Check for finish_reason to emit terminal events AFTER content
        if candidate0:
            finish_reason = candidate0.get("finish_reason")
            if finish_reason:
                if finish_reason == "MAX_TOKENS":
                    results.append(
                        ResponseIncompleteEvent(
                            response=self._convert_result(event)
                        )
                    )
                elif finish_reason in ("SAFETY", "BLOCKED", "RECITATION"):
                    results.append(
                        ResponseFailedEvent(
                            response=self._convert_result(event)
                        )
                    )
                else:
                    # STOP or other normal completion
                    results.append(
                        ResponseCompletedEvent(
                            response=self._convert_result(event)
                        )
                    )

        return results

    def _convert_result(self, response: Any) -> TextGenerationResult:
        r: dict[str, Any] = (
            response.model_dump()
            if hasattr(response, "model_dump")
            else response
        )

        output_items: list[OutputItem] = []
        message_content: list[OutputMessageContent] = []
        reasoning_content: list[OutputReasoningContentText] = []

        # Extract first candidate content parts.
        candidates = r.get("candidates")
        candidate0: dict[str, Any] | None = None
        if isinstance(candidates, list) and candidates:
            c0 = candidates[0]
            if isinstance(c0, dict):
                candidate0 = c0

        if candidate0:
            content = candidate0.get("content")
            if isinstance(content, dict):
                parts = content.get("parts")
                if isinstance(parts, list):
                    for part in parts:
                        if not isinstance(part, dict):
                            continue
                        # Text parts - "thought" indicates reasoning text
                        txt = part.get("text")
                        if isinstance(txt, str):
                            if part.get("thought"):
                                reasoning_content.append(
                                    OutputReasoningContentText(text=txt)
                                )
                            else:
                                message_content.append(OutputText(text=txt))
                        # Function call parts
                        fc = part.get("function_call")
                        if isinstance(fc, dict) and fc.get("name"):
                            output_items.append(
                                FunctionCall(
                                    name=fc.get("name") or "",
                                    arguments=fc.get("args"),
                                    call_id=fc.get("id"),
                                    thought_signature=part.get(
                                        "thought_signature"
                                    ),
                                )
                            )

        # Add reasoning as separate output item if present
        if reasoning_content:
            output_items.insert(
                0,
                OutputReasoning(
                    content=reasoning_content,
                    status="completed",
                ),
            )

        # Add assistant message if any content collected
        if message_content:
            output_items.insert(
                0,
                OutputMessage(
                    role="assistant",
                    content=message_content,
                    status="completed",
                ),
            )

        # Usage mapping from usage_metadata
        usage_obj: Usage | None = None
        usage = r.get("usage_metadata")
        if isinstance(usage, dict):
            input_tokens = int(usage.get("prompt_token_count") or 0)
            output_tokens = int(usage.get("candidates_token_count") or 0)
            total_tokens = int(usage.get("total_token_count") or 0)

            def _modality_details(items: Any) -> dict[str, int] | None:
                if not isinstance(items, list) or not items:
                    return None
                details: dict[str, int] = {}
                for it in items:
                    if isinstance(it, dict):
                        mod = it.get("modality")
                        cnt = it.get("token_count")
                        if mod is not None and isinstance(cnt, int):
                            details[str(mod)] = cnt
                return details or None

            usage_obj = Usage(
                input_tokens=input_tokens,
                # input_tokens_details=_modality_details(
                #    usage.get("prompt_tokens_details")
                # ),
                output_tokens=output_tokens,
                output_tokens_details={
                    "reasoning_tokens": int(
                        usage.get("thoughts_token_count") or 0
                    )
                },
                total_tokens=total_tokens,
            )

        # Error handling
        error_obj: ErrorDetail | None = None
        # Check for blocked content or other errors
        if candidate0:
            finish_reason = candidate0.get("finish_reason")
            if finish_reason in ("SAFETY", "BLOCKED", "RECITATION"):
                error_obj = ErrorDetail(
                    code=finish_reason.lower(),
                    message=f"Response blocked due to {finish_reason}",
                )
            # Check for content filter results
            safety_ratings = candidate0.get("safety_ratings")
            if isinstance(safety_ratings, list):
                for rating in safety_ratings:
                    if isinstance(rating, dict) and rating.get("blocked"):
                        error_obj = ErrorDetail(
                            code="content_filter",
                            message=f"Blocked by {rating.get('category')}",
                        )
                        break

        # Determine status
        status: Literal[
            "completed",
            "failed",
            "in_progress",
            "cancelled",
            "queued",
            "incomplete",
        ] = "completed"
        if candidate0:
            finish_reason = candidate0.get("finish_reason")
            if finish_reason == "MAX_TOKENS":
                status = "incomplete"
            elif finish_reason in ("SAFETY", "BLOCKED", "RECITATION"):
                status = "failed"
        if error_obj:
            status = "failed"

        # created_at: convert datetime to epoch seconds if present
        created_at: int | None = None
        create_time = r.get("create_time")
        try:
            if create_time is not None and hasattr(create_time, "timestamp"):
                created_at = int(create_time.timestamp())
        except Exception:
            created_at = None

        result = TextGenerationResult(
            id=r.get("response_id"),
            model=r.get("model_version"),
            created_at=created_at,
            status=status,
            error=error_obj,
            output=output_items or None,
            usage=usage_obj,
        )
        return result

    def _convert_generate_args(
        self,
        input: str | list[dict[str, Any] | InputItem],
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        text: dict | ResponseText | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        args: dict[str, Any] = {}
        args["model"] = model or self.model

        contents: list[dict[str, Any]] = []

        def _convert_content_part(c: dict[str, Any]) -> dict[str, Any] | None:
            """Convert a single content part to Google format."""
            ct = c.get("type")
            if ct == "input_text":
                return {"text": c.get("text")}
            if ct == "output_text":
                # Handle output text from previous assistant messages
                return {"text": c.get("text")}
            if ct == "input_image":
                img = c.get("image")
                if not isinstance(img, dict):
                    return None
                media_type = (
                    img.get("media_type") or "application/octet-stream"
                )
                content = img.get("content")
                source = img.get("source")
                if isinstance(content, (bytes, bytearray)):
                    return {
                        "inline_data": {
                            "data": bytes(content),
                            "mime_type": media_type,
                        }
                    }
                elif isinstance(content, str):
                    return {
                        "inline_data": {
                            "data": base64.b64decode(content),
                            "mime_type": media_type,
                        }
                    }
                if isinstance(source, str):
                    return {
                        "file_data": {
                            "file_uri": source,
                            "mime_type": media_type,
                        }
                    }
            if ct == "input_file":
                f = c.get("file")
                if isinstance(f, dict):
                    content = f.get("content")
                    mime_type = (
                        f.get("media_type") or "application/octet-stream"
                    )
                    if isinstance(content, (bytes, bytearray)):
                        return {
                            "inline_data": {
                                "data": bytes(content),
                                "mime_type": mime_type,
                            }
                        }
                    elif isinstance(content, str):
                        # Treat string content as base64-encoded
                        return {
                            "inline_data": {
                                "data": base64.b64decode(content),
                                "mime_type": mime_type,
                            }
                        }
                    source = f.get("source")
                    if isinstance(source, str):
                        return {
                            "file_data": {
                                "file_uri": source,
                                "mime_type": mime_type,
                            }
                        }
            return None

        def _convert_content_to_parts(
            raw_content: str | list | None,
        ) -> list[dict[str, Any]]:
            """Convert message content to Google parts list."""
            if isinstance(raw_content, str):
                return [{"text": raw_content}]
            if isinstance(raw_content, list):
                parts = []
                for c in raw_content:
                    if isinstance(c, dict):
                        part = _convert_content_part(c)
                        if part:
                            parts.append(part)
                return parts
            return []

        def _convert_input_item(msg: InputItem | dict[str, Any]) -> None:
            m: dict[str, Any] = msg if isinstance(msg, dict) else msg.to_dict()
            item_type = m.get("type")

            if item_type == "function_call":
                fc_dict: dict[str, Any] = {
                    "name": m.get("name"),
                    "args": m.get("arguments") or {},
                }
                # Build the part with function_call
                part: dict[str, Any] = {"function_call": fc_dict}
                thought_sig = m.get("thought_signature")
                if thought_sig:
                    part["thought_signature"] = thought_sig
                contents.append(
                    {
                        "role": "model",
                        "parts": [part],
                    }
                )
                return

            if item_type == "function_call_output":
                contents.append(
                    {
                        "role": "user",
                        "parts": [
                            {
                                "function_response": {
                                    "name": m.get("name") or m.get("call_id"),
                                    "response": {"output": m.get("output")},
                                }
                            }
                        ],
                    }
                )
                return

            # Handle OutputReasoning (skip - not sent back to Google)
            if item_type == "reasoning":
                return

            if item_type != "message":
                return

            role = m.get("role")
            # Map roles to Google format
            if role == "assistant":
                g_role = "model"
            elif role in ("system", "developer"):
                g_role = "user"
            else:
                g_role = "user"

            parts = _convert_content_to_parts(m.get("content"))
            contents.append({"role": g_role, "parts": parts})

        # Input may be a raw string or a sequence of InputItems.
        if isinstance(input, str):
            contents.append(
                {"role": "user", "parts": [types.Part.from_text(text=input)]}
            )
        else:
            for it in input:
                _convert_input_item(it)

        args["contents"] = contents
        config: dict[str, Any] = {}
        if temperature:
            config["temperature"] = temperature
        if top_p:
            config["top_p"] = top_p
        if max_output_tokens:
            config["max_output_tokens"] = max_output_tokens
        if text:
            config["response_modalities"] = ["TEXT"]
            text_dict = text if isinstance(text, dict) else text.to_dict()
            fmt = text_dict.get("format")
            if isinstance(fmt, dict) and fmt.get("type") == "json_schema":
                config["response_mime_type"] = "application/json"
                schema = fmt.get("schema")
                if schema:
                    config["response_schema"] = schema
        if tools:
            config["tools"] = self._convert_tools(tools)
        if tool_choice:
            config["tool_config"] = self._convert_tool_choice(tool_choice)
        if instructions:
            config["system_instruction"] = instructions
        if reasoning:
            if isinstance(reasoning, dict):
                effort = reasoning.get("effort", "low")
            elif isinstance(reasoning, Reasoning):
                effort = reasoning.effort
            else:
                effort = "low"
            thinking_level_map: dict[str, types.ThinkingLevel] = {
                "none": types.ThinkingLevel.MINIMAL,
                "low": types.ThinkingLevel.LOW,
                "medium": types.ThinkingLevel.MEDIUM,
                "high": types.ThinkingLevel.HIGH,
            }
            config["thinking_config"] = types.ThinkingConfig(
                include_thoughts=True,
                thinking_level=thinking_level_map.get(
                    effort, types.ThinkingLevel.MINIMAL
                ),
            )
        if config:
            args["config"] = config
        if nconfig:
            args.update(nconfig)
        return args

    def _convert_function(self, fn: dict) -> types.FunctionDeclaration:
        decl = types.FunctionDeclaration(
            name=fn.get("name"),
            description=fn.get("description"),
            parameters_json_schema=fn.get("parameters"),
        )
        return decl

    def _convert_tools(self, tools: list[dict | Tool]) -> list[types.Tool]:
        function_declarations: list[types.FunctionDeclaration] = []
        google_search = None
        for t in tools:
            if isinstance(t, dict):
                tool_type = t.get("type", "function")
                tool = t
            else:
                tool_type = t.type
                tool = t.to_dict()
            if tool_type == "function":
                decl = self._convert_function(tool)
                function_declarations.append(decl)
            elif tool_type == "web_search":
                google_search = types.GoogleSearch()
            else:
                raise BadRequestError(f"Unsupported tool type: {tool_type}")
        return [
            types.Tool(
                function_declarations=function_declarations or None,
                google_search=google_search,
            )
        ]

    def _convert_tool_choice(
        self,
        choice: dict | ToolChoice | None,
    ) -> types.ToolConfig | None:
        if choice is None:
            return None

        if isinstance(choice, ToolChoiceFunction) or isinstance(
            choice, AllowedTools
        ):
            inp = choice.to_dict()
        else:
            inp = choice

        choice_mode_map: dict = {
            "none": types.FunctionCallingConfigMode.NONE,
            "auto": types.FunctionCallingConfigMode.AUTO,
            "required": types.FunctionCallingConfigMode.ANY,
        }
        if inp in ("none", "auto", "required"):
            return types.ToolConfig(
                function_calling_config=types.FunctionCallingConfig(
                    mode=choice_mode_map[inp]
                )
            )
        if isinstance(inp, dict):
            if inp.get("type") == "function":
                name = inp.get("name")
                if not name:
                    raise BadRequestError(
                        "Function tool_choice must specify a name"
                    )
                return types.ToolConfig(
                    function_calling_config=types.FunctionCallingConfig(
                        mode=types.FunctionCallingConfigMode.ANY,
                        allowed_function_names=[name],
                    )
                )
            elif inp.get("type") == "allowed_tools":
                mode = inp.get("mode")
                allowed = inp.get("tools")
                allowed_function_names = (
                    [t.get("name") for t in allowed]
                    if isinstance(allowed, list)
                    else None
                )
                return types.ToolConfig(
                    function_calling_config=types.FunctionCallingConfig(
                        mode=choice_mode_map[mode],
                        allowed_function_names=allowed_function_names,
                    )
                )

        raise BadRequestError(f"Unsupported tool_choice: {type(choice)}")



from typing import Any

from common.secrets import get_secrets

from x8.ai.text_generation import TextGeneration
from x8.ai.video_generation import VideoGeneration

secrets = get_secrets()


class VideoGenerationProvider:
    OPENAI = "openai"
    GOOGLE = "google"


provider_types: dict[str, str] = {
    VideoGenerationProvider.OPENAI: "openai",
    VideoGenerationProvider.GOOGLE: "google",
}


provider_parameters: dict[str, dict[str, Any]] = {
    VideoGenerationProvider.OPENAI: {
        "api_key": secrets["ai-openai-api-key"],
    },
    VideoGenerationProvider.GOOGLE: {},
}


def get_component(provider_type: str):
    parameters = provider_parameters[provider_type]
    component = VideoGeneration(
        __unpack__=True,
        __provider__=dict(
            type=provider_types[provider_type],
            parameters=parameters,
        ),
    )
    return component


def get_judge():
    component = TextGeneration(
        __unpack__=True,
        __provider__=dict(
            type="openai",
            parameters={
                "api_key": secrets["OPENAI_API_KEY"],
            },
        ),
    )
    return component



from __future__ import annotations

from typing import Any, AsyncIterator, Iterator, Literal

from x8.core import DataModel

# ── Tool definitions ──────────────────────────────────────────────────


class ToolParameter(DataModel):
    """JSON-Schema style parameter definition for a tool."""

    type: str = "object"
    properties: dict[str, Any] | None = None
    required: list[str] | None = None


class Tool(DataModel):
    """A tool that the agent can invoke."""

    name: str
    description: str | None = None
    parameters: dict[str, Any] | ToolParameter | None = None


# ── MCP server definitions ───────────────────────────────────────────


class MCPStdioServer(DataModel):
    """An MCP server that runs as a subprocess (stdio transport)."""

    type: Literal["stdio"] = "stdio"
    command: str
    args: list[str] | None = None
    env: dict[str, str] | None = None


class MCPServer(DataModel):
    """Wrapper that carries a name + config for an MCP server."""

    name: str
    server: MCPStdioServer | dict[str, Any]


# ── Content blocks ───────────────────────────────────────────────────


class TextBlock(DataModel):
    """A block of text produced by the agent."""

    type: Literal["text"] = "text"
    text: str


class ToolUseBlock(DataModel):
    """A tool invocation made by the agent."""

    type: Literal["tool_use"] = "tool_use"
    tool_name: str
    tool_input: dict[str, Any] | None = None
    call_id: str | None = None


class ToolResultBlock(DataModel):
    """The result returned after a tool executes."""

    type: Literal["tool_result"] = "tool_result"
    tool_name: str | None = None
    call_id: str | None = None
    output: str | dict[str, Any] | None = None
    is_error: bool = False


ContentBlock = TextBlock | ToolUseBlock | ToolResultBlock


# ── Messages ─────────────────────────────────────────────────────────


class Message(DataModel):
    """A single message in the agent conversation."""

    role: Literal["user", "assistant", "system", "tool"] = "assistant"
    content: str | list[ContentBlock] | None = None


# ── Usage ────────────────────────────────────────────────────────────


class Usage(DataModel):
    """Token usage information for an agent run."""

    input_tokens: int | None = None
    output_tokens: int | None = None
    total_tokens: int | None = None


# ── Agent run result ─────────────────────────────────────────────────


class AgentResult(DataModel):
    """The final result of an agent run."""

    id: str | None = None
    model: str | None = None
    status: (
        Literal[
            "completed",
            "failed",
            "in_progress",
            "cancelled",
            "queued",
            "incomplete",
        ]
        | None
    ) = "completed"
    output: list[Message] | None = None
    usage: Usage | None = None
    error: str | None = None


# ── Streaming events ─────────────────────────────────────────────────


class StreamEvent(DataModel):
    """Base streaming event."""

    type: str


class TextDeltaEvent(StreamEvent):
    """Incremental text produced by the agent."""

    type: Literal["text_delta"] = "text_delta"
    delta: str


class TextDoneEvent(StreamEvent):
    """A text block is complete."""

    type: Literal["text_done"] = "text_done"
    text: str


class ToolCallEvent(StreamEvent):
    """The agent is invoking a tool."""

    type: Literal["tool_call"] = "tool_call"
    tool_name: str
    tool_input: dict[str, Any] | None = None
    call_id: str | None = None


class ToolResultEvent(StreamEvent):
    """A tool execution completed."""

    type: Literal["tool_result"] = "tool_result"
    tool_name: str | None = None
    call_id: str | None = None
    output: str | dict[str, Any] | None = None
    is_error: bool = False


class AgentCompletedEvent(StreamEvent):
    """Agent run completed."""

    type: Literal["completed"] = "completed"
    result: AgentResult


class ErrorEvent(StreamEvent):
    """An error occurred during the agent run."""

    type: Literal["error"] = "error"
    code: str | None = None
    message: str | None = None


AgentStreamEvent = (
    TextDeltaEvent
    | TextDoneEvent
    | ToolCallEvent
    | ToolResultEvent
    | AgentCompletedEvent
    | ErrorEvent
)


# ── Session ──────────────────────────────────────────────────────────


class SessionConfig(DataModel):
    """Configuration for creating an agent session."""

    model: str | None = None
    instructions: str | None = None
    tools: list[Tool] | None = None
    mcp_servers: list[MCPServer | MCPStdioServer] | None = None
    allowed_tools: list[str] | None = None
    max_turns: int | None = None
    cwd: str | None = None
    nconfig: dict[str, Any] | None = None


class AgentSession:
    """A persistent, multi-turn agent session.

    Sessions keep the underlying client alive across multiple
    ``send`` / ``asend`` calls, enabling true conversational
    interaction with the agent.

    Use as a context manager (sync or async)::

        async with agent.acreate_session() as session:
            r1 = await session.asend("Hello")
            r2 = await session.asend("Follow up")

        with agent.create_session() as session:
            r1 = session.send("Hello")
            r2 = session.send("Follow up")
    """

    config: SessionConfig
    history: list[Message]

    def __init__(self, config: SessionConfig | None = None):
        self.config = config or SessionConfig()
        self.history = []

    # ── sync API ──────────────────────────────────────────────────

    def send(
        self,
        prompt: str,
        *,
        stream: bool | None = None,
        **kwargs: Any,
    ) -> AgentResult | Iterator[AgentStreamEvent]:
        """Send a message and get a response."""
        raise NotImplementedError

    def close(self) -> None:
        """Close the session and release resources."""
        raise NotImplementedError

    # ── async API ────────────────────────────────────────────────

    async def asend(
        self,
        prompt: str,
        *,
        stream: bool | None = None,
        **kwargs: Any,
    ) -> AgentResult | AsyncIterator[AgentStreamEvent]:
        """Send a message and get a response (async)."""
        raise NotImplementedError

    async def aclose(self) -> None:
        """Close the session and release resources (async)."""
        raise NotImplementedError

    # ── context managers ─────────────────────────────────────────

    def __enter__(self) -> "AgentSession":
        return self

    def __exit__(self, *exc) -> None:
        self.close()

    async def __aenter__(self) -> "AgentSession":
        return self

    async def __aexit__(self, *exc) -> None:
        await self.aclose()



from __future__ import annotations

from typing import Any, Literal

from x8.core import DataModel

MessageChannel = Literal[
    "sms",
    "whatsapp",
    "telegram",
    "messenger",
    "rcs",
    "other",
]


class MessageMedia(DataModel):
    url: str
    content_type: str | None = None
    caption: str | None = None


class MessageContent(DataModel):
    text: str | None = None
    media: list[MessageMedia] | None = None


class MessageSendResult(DataModel):
    id: str
    status: str | None = None
    channel: MessageChannel | None = None
    to: str | None = None
    from_: str | None = None
    provider: str | None = None
    price: str | None = None
    currency: str | None = None
    info: dict[str, Any] = dict()


class VerificationStartResult(DataModel):
    id: str
    status: str | None = None
    channel: MessageChannel | None = None
    to: str | None = None
    provider: str | None = None
    info: dict[str, Any] = dict()


class VerificationCheckResult(DataModel):
    id: str
    status: str | None = None
    approved: bool | None = None
    valid: bool | None = None
    to: str | None = None
    provider: str | None = None
    info: dict[str, Any] = dict()



from ._models import EmailAttachment, EmailContent, EmailSendResult
from .component import EmailService

__all__ = [
    "EmailAttachment",
    "EmailContent",
    "EmailSendResult",
    "EmailService",
]



from typing import Any

from common.secrets import get_secrets

from x8.ai.text_generation import TextGeneration

secrets = get_secrets()


class TextGenerationProvider:
    OPENAI = "openai"
    OPENAI_LEGACY = "openai_legacy"
    GOOGLE = "google"
    ANTHROPIC = "anthropic"
    XAI = "xai"
    DEEPSEEK = "deepseek"
    OLLAMA = "ollama"
    GROQ = "groq"
    GROQ_LLAMA = "groq_llama"
    MISTRAL = "mistral"
    COHERE = "cohere"
    COHERE_VISION = "cohere_vision"
    AZURE_OPENAI = "azure_openai"
    AMAZON_BEDROCK = "amazon_bedrock"
    TOGETHER = "together"
    FIREWORKS = "fireworks"
    REPLICATE = "replicate"


provider_types: dict[str, str] = {
    TextGenerationProvider.OPENAI: "openai",
    TextGenerationProvider.OPENAI_LEGACY: "openai_legacy",
    TextGenerationProvider.GOOGLE: "google",
    TextGenerationProvider.ANTHROPIC: "anthropic",
    TextGenerationProvider.XAI: "xai",
    TextGenerationProvider.DEEPSEEK: "deepseek",
    TextGenerationProvider.OLLAMA: "ollama",
    TextGenerationProvider.GROQ: "groq",
    TextGenerationProvider.GROQ_LLAMA: "groq",
    TextGenerationProvider.MISTRAL: "mistral",
    TextGenerationProvider.COHERE: "cohere",
    TextGenerationProvider.COHERE_VISION: "cohere",
    TextGenerationProvider.AZURE_OPENAI: "azure_openai",
    TextGenerationProvider.AMAZON_BEDROCK: "amazon_bedrock",
    TextGenerationProvider.TOGETHER: "together",
    TextGenerationProvider.FIREWORKS: "fireworks",
    TextGenerationProvider.REPLICATE: "replicate",
}


provider_parameters: dict[str, dict[str, Any]] = {
    TextGenerationProvider.OPENAI: {
        "api_key": secrets["OPENAI_API_KEY"],
    },
    TextGenerationProvider.OPENAI_LEGACY: {
        "api_key": secrets["OPENAI_API_KEY"],
    },
    TextGenerationProvider.GOOGLE: {},
    TextGenerationProvider.ANTHROPIC: {
        "api_key": secrets["ANTHROPIC_API_KEY"],
    },
    TextGenerationProvider.XAI: {
        "api_key": secrets["XAI_API_KEY"],
    },
    TextGenerationProvider.DEEPSEEK: {
        "api_key": secrets["DEEPSEEK_API_KEY"],
    },
    TextGenerationProvider.OLLAMA: {},
    TextGenerationProvider.GROQ: {
        "api_key": secrets["GROQ_API_KEY"],
    },
    TextGenerationProvider.GROQ_LLAMA: {
        "api_key": secrets["GROQ_API_KEY"],
        "model": "meta-llama/llama-4-scout-17b-16e-instruct",
    },
    TextGenerationProvider.MISTRAL: {
        "api_key": secrets["MISTRAL_API_KEY"],
    },
    TextGenerationProvider.COHERE: {
        "api_key": secrets["COHERE_API_KEY"],
    },
    TextGenerationProvider.COHERE_VISION: {
        "api_key": secrets["COHERE_API_KEY"],
        "model": "command-a-vision-07-2025",
    },
    TextGenerationProvider.AZURE_OPENAI: {
        "api_version": "2024-12-01-preview",
        "azure_endpoint": secrets["AZURE_OPENAI_ENDPOINT"],  # noqa
        "model": "model-router",
        "api_key": secrets["AZURE_OPENAI_API_KEY"],
    },
    TextGenerationProvider.AMAZON_BEDROCK: {
        "model": "us.amazon.nova-lite-v1:0",
        "region": "us-east-1",
    },
    TextGenerationProvider.TOGETHER: {
        "api_key": secrets["TOGETHER_API_KEY"],
        "model": "Qwen/Qwen3-VL-32B-Instruct",
    },
    TextGenerationProvider.FIREWORKS: {
        "api_key": secrets["FIREWORKS_API_KEY"],
        "model": "accounts/fireworks/models/qwen3-vl-235b-a22b-thinking",
    },
    TextGenerationProvider.REPLICATE: {
        "api_key": secrets["REPLICATE_API_KEY"],
        "model": "google/gemini-3-pro",
    },
}


def get_component(provider_type: str):
    parameters = provider_parameters[provider_type]
    component = TextGeneration(
        __unpack__=True,
        __provider__=dict(
            type=provider_types[provider_type],
            parameters=parameters,
        ),
    )
    return component
